"""
A collection of classes used to interrogate MinKNOW
"""

import datetime
import json
import os
import sys
import time
import logging

from urllib.parse import urlparse

import requests
from ws4py.client.threadedclient import WebSocketClient

from minFQ.channelmaps import chanlookup_type
from minFQ.minotourapi import MinotourAPI as MinotourAPINew


log = logging.getLogger(__name__)

global channel_data  # TODO does it need to be global?

channel_data=dict()

def startstop(command,minION):
    if self.args.GUI:
        self.args.minKNOWmessage = "Sending start stop message."
    if OPER == "osx":
        p = os.popen('/Applications/MinKNOW.app/Contents/Resources/bin/mk_manager_client -i ' + minION + ' --' + command,"r")
        while 1:
            line = p.readline()
            if not line: break
    elif OPER == "windows":
        p = os.popen('C:\\grouper\\binaries\\bin\\mk_manager_client.exe -i ' + minION + ' --' + command, "r")
        while 1:
            line = p.readline()
            if not line: break
    elif OPER == "linux":
        p = os.popen('/opt/ONT/MinKNOW/bin/mk_manager_client -i' + minION + ' --' + command, "r")
        while 1:
            line = p.readline()
            if not line: break
    else:
        print ("!!!!!!!!!!!!!! Sorry - cannot recognise your operating system.")

def send_message_port(message,ipadd,port):
    message_to_send = \
        '{"id":"1", "method":"user_message","params":{"content":"%s"}}' \
        % message
    results=""
    try:
        results = execute_command_as_string(message_to_send, ipadd, port)
    except Exception as err:
        print ("message send fail", err)
    return results

def startrun(script,ipadd,port):
    try:
        startruncustom = \
            '{"id":1, "method":"start_script","params":{"name":"' \
            + script + '"}}'
        startresult = \
            execute_command_as_string(startruncustom,
                ipadd, port)
        startrunmessage = 'minoTour sent a remote run start command.'
        startresultmessage = send_message_port(startrunmessage, ipadd, port)
    except Exception as err:
        print >> sys.stderr, err

def stoprun_message(port,ipadd,message):
    try:
        stopresult = execute_command_as_string(commands('stoprun'),
                ipadd, port)
        stopprotocolresult = \
            execute_command_as_string(commands('stopprotocol'), ipadd,
                port)
        stoprunmessage = 'minoTour sent a remote run stop command for %s.' % (message)
        stopresultmessage = send_message_port(stoprunmessage,port)
    except Exception as err:
        print >> sys.stderr, err

def send_custom_message(port,message):
    try:
        custommessage = "minoTour: %s" % (message)
        custommessageresult=send_message_port(custommessage,port)
    except Exception as err:
        print >> sys.stderr, err

def stoprun(ipadd,port):
    try:
        stopresult = execute_command_as_string(commands('stoprun'),
                ipadd, port)
        stopprotocolresult = \
            execute_command_as_string(commands('stopprotocol'), ipadd,
                port)
        stoprunmessage = 'minoTour sent a remote run stop command.'
        stopresultmessage = send_message_port(stoprunmessage,ipadd,port)
    except Exception as err:
        print >> sys.stderr, err


def renameflowcell(name,ipadd,port):
    try:
        set_flowcell_id = \
            '{"id":1,"method":"set_engine_state","params":{"state_id":"flow_cell_id","value":"'+name+'"}}'
        set_sample_id_result = \
            execute_command_as_string(set_flowcell_id,ipadd,port)
        startresultmessage = \
            send_message_port('minoTour renamed the flowcell to ' + name,ipadd,port)
    except Exception as err:
        print >> sys.stderr, err

def renamerun(name,ipadd,port):
    try:
        set_sample_id = \
            '{"id":"1","method":"set_engine_state","params":{"state_id":"sample_id","value":"' \
            + name + '"}}'
        set_sample_id_result = \
            execute_command_as_string(set_sample_id, ipadd,
                port)
        startresultmessage = \
            send_message_port('minoTour renamed the run to ' + name,ipadd,port)
    except Exception as err:
        print >> sys.stderr, err


def execute_command_as_string(data, host=None, port=None):
    """
    Notes for requests
    header = {'Authorization':'Token e45c142b457121278f9b67d713285a7e10382b36', 'Content-Type':'application/json'}
    r=requests.post(args.full_host+'api/v1/runs/', headers=header, json={"run_name": "20170612_1630_matt", "run_id": "hj78yy9o-217e-4335-9451-66a7288a9dd5", "barcode": "barcode10"})
    r.text #returns result
    """
    host_name = host
    port_number = port
    url = 'http://%s:%s%s' % (str(host_name), str(port_number), '/jsonrpc')
    r=requests.post(url, data=data,headers={'Content-Length': str(len(data)),'Content-Type': 'application/json'})
    try:
        json_respond = json.loads(r.text)
        return json_respond
    except Exception as err:
        print (err)
        err_string = \
            'Fail to initialise mincontrol. Likely reasons include minKNOW not running, the wrong IP address for the minKNOW server or firewall issues.'
    return


def commands(command):
    return {
        'getstaticdata' : '{"id":1,"method":"get_static_data","params":null}',
        'initialization_status' : '{"params": "null", "id": 5, "method": "initialization_status"}',
        'get_analysis_configuration' : '{"id":1,"method":"get_analysis_configuration","params":null}',
        'initialiseminion' : '{"params": {"command": "init_main_board", "parameters": []}, "id": 0, "method": "board_command_ex"}',
        'shutdownminion' : '{"params": {"state_id": "status", "value": "stop"}, "id": 0, "method": "set_engine_state"}',
        'startmessagenew' : '{"id":"1", "method":"user_message","params":{"content":"minoTour is now interacting with your run. This is done at your own risk. To stop minoTour interaction with minKnow disable upload of read data to minoTour."}}',
        'status':'{"id":"1", "method":"get_engine_state","params":{"state_id":"status"}}',
        'dataset':'{"id":"1", "method":"get_engine_state","params":{"state_id":"data_set"}}',
        'startrun' :'{"id":"1", "method":"start_script","params":{"name":"MAP_Lambda_Burn_In_Run_SQK_MAP005.py"}}',
        'stoprun' : '{"id":"1", "method":"stop_experiment","params":"null"}',
        'stopprotocol' : '{"id":"1", "method":"stop_script","params":{"name":"MAP_48Hr_Sequencing_Run.py"}}',
        'biasvoltageget' : '{"id":"1","method":"board_command_ex","params":{"command":"get_bias_voltage"}}',
        'bias_voltage_gain' : '{"id":"1","method":"get_engine_state","params":{"state_id":"bias_voltage_gain"}}',
        'bias_voltage_set' :'{"id":"1","method":"board_command_ex","params":{"command":"set_bias_voltage","parameters":"-120"}}',
        'machine_id' :'{"id":"1","method":"get_engine_state","params":{"state_id":"machine_id"}}',
        'machine_name' :'{"id":"1","method":"get_engine_state","params":{"state_id":"machine_name"}}',
        'sample_id' :'{"id":"1","method":"get_engine_state","params":{"state_id":"sample_id"}}',
        'flow_cell_id' : '{"id":"1","method":"get_engine_state","params":{"state_id":"flow_cell_id"}}',
        'user_error' :'{"id":"1","method":"get_engine_state","params":{"state_id":"user_error"}}',
        'sequenced_res' :'{"id":"1","method":"get_engine_state","params":{"state_id":"sequenced"}}',
        'yield_res' : '{"id":"1","method":"get_engine_state","params":{"state_id":"yield"}}',
        'current_script' : '{"id":"1","method":"get_engine_state","params":{"state_id":"current_script"}}',
        'get_scripts' : '{"id":"1", "method":"get_script_info_list","params":{"state_id":"status"}}',
        'disk_space' : '{"id":1,"method":"get_disk_space_info","params":null}',
        'sinc_delay' : '{"id":1,"method":"sinc_delay","params":null}',
        'get_seq_metrics' : '{"id":1,"method": "get_seq_metrics","params":null}',
        'get_tracking_id': '{"id":1,"method": "get_tracking_id","params":null}',
        'get_statistics': '{"id":1,"method": "get_statistics","params":null}',
        'get_engine_states' : '{"id":1,"method": "get_engine_states","params":null}',

    }[command]


class MessagesClient(WebSocketClient):
    def __init__(self, connectip,args,minIONdict):
        super(MessagesClient, self).__init__(connectip)
        self.detailsdict=dict()
        self.daemon=True
        self.args=args
        self.minIONdict=minIONdict

    def init_minion(self,minion):
        self.minion=minion

    def opened(self):
        print ("MessagesClient Opened!")

    def closed(self, code, reason="MessagesClient disconnected for some reason"):
        print ("socket",self.sock.fileno())
        print ("Closed down", code, reason)

    def received_message(self,m):
        if not m.is_binary:
            if hasattr(self, 'minion'):
                try:
                    if self.minion not in self.minIONdict.keys():
                        self.minIONdict[self.minion]=dict()
                    if "messages" not in self.minIONdict[self.minion].keys():
                        self.minIONdict[self.minion]["messages"]=[]
                    self.minIONdict[self.minion]["APIHelp"].update_message(json.loads(str(m)))
                except Exception as err:
                    print (err)
                    pass
            else:
                print ("Still waiting to configure")

def get_run_scripts(ipadd,port):
    get_scripts = \
        '{"id":"1", "method":"get_script_info_list","params":null}'
    results = execute_command_as_string(get_scripts, ipadd, port)
    return results['result']


def process_tracked_yield(minIONdict):
    """
    This function maintains a history of all yield values over the time of a run - it can be configured to store yields
    at given intervals. It is used by the main web page to provide a history of cumulative yield.
    """
    for minion in minIONdict:
        if "detailsdata" in minIONdict[minion]:
            if "livedata" in minIONdict[minion]:
                if "yield_res" in minIONdict[minion]["livedata"]:
                    if "result" in minIONdict[minion]["livedata"]["yield_res"]:
                        try:
                            asictemp = float(minIONdict[minion]["detailsdata"]["engine_states"]["minion_asic_temperature"])
                        except:
                            asictemp = 0
                        try:
                            heatsinktemp = float(minIONdict[minion]["detailsdata"]["engine_states"]["minion_heatsink_temperature"])
                        except:
                            heatsinktemp = 0
                        if minIONdict[minion]["livedata"]["bias_voltage_gain"]["result"] != "null":
                            biasvoltage = int(minIONdict[minion]["livedata"]["bias_voltage_gain"]["result"])
                        else:
                            biasvoltage = 0
                        if minIONdict[minion]["livedata"]["biasvoltageget"]["result"] != "null":
                            voltage_val = int(minIONdict[minion]["livedata"]["biasvoltageget"]["result"]["bias_voltage"])
                        else:
                            voltage_val = 0
                        voltage_value = voltage_val * biasvoltage
                        try:
                            if minIONdict[minion]["livedata"]["yield_res"]["result"] != "null":
                                yieldval = int(minIONdict[minion]["livedata"]["yield_res"]["result"])
                            else:
                                yieldval =  minIONdict[minion]["detailsdata"]["statistics"]["read_event_count"]
                        except:
                            yieldval=0
                        try:
                            meanratio = minIONdict[minion]["detailsdata"]["meanratio"]
                            openpore = minIONdict[minion]["detailsdata"]["openpore"]
                            instrand = minIONdict[minion]["detailsdata"]["instrand"]
                        except:
                            meanratio = 0
                            openpore = 0
                            instrand = 0
                        if "yield_history" not in minIONdict[minion]:
                            minIONdict[minion]["yield_history"]=[]
                        if "temp_history" not in minIONdict[minion]:
                            minIONdict[minion]["temp_history"]=dict()
                        if "pore_history" not in minIONdict[minion]:
                            minIONdict[minion]["pore_history"]=dict()
                        if "meanratio_history" not in minIONdict[minion]["pore_history"]:
                            minIONdict[minion]["pore_history"]["meanratio_history"]=[]
                            minIONdict[minion]["pore_history"]["openpore_history"] = []
                            minIONdict[minion]["pore_history"]["instrand_history"] = []
                        if "strand" not in minIONdict[minion]["pore_history"]:
                            minIONdict[minion]["pore_history"]["strand"]=[]
                            minIONdict[minion]["pore_history"]["percent"]=[]
                            minIONdict[minion]["pore_history"]["single"]=[]
                        if "asictemp" not in minIONdict[minion]["temp_history"]:
                            minIONdict[minion]["temp_history"]["asictemp"]=[]
                            minIONdict[minion]["temp_history"]["heatsinktemp"]=[]
                            minIONdict[minion]["temp_history"]["voltage"]=[]
                        strand = 0
                        single_pore = 0
                        if "simplesummary" in minIONdict[minion]:
                            if "strand" in minIONdict[minion]["simplesummary"].keys():
                                strand = float(minIONdict[minion]["simplesummary"]["strand"])
                            if "single_pore" in minIONdict[minion]["simplesummary"].keys():
                                single_pore = float(minIONdict[minion]["simplesummary"]["single_pore"])
                            elif "good_single" in minIONdict[minion]["simplesummary"].keys():
                                single_pore = float(minIONdict[minion]["simplesummary"]["good_single"])
                        if (strand+single_pore)> 0:
                            percent = (strand/(strand+single_pore)*100)
                        else:
                            percent = 0
                        if len(minIONdict[minion]["yield_history"]) >1:
                            if yieldval > minIONdict[minion]["yield_history"][-1][1]:
                                minIONdict[minion]["yield_history"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,yieldval))
                            minIONdict[minion]["temp_history"]["asictemp"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,asictemp))
                            minIONdict[minion]["temp_history"]["voltage"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,voltage_value))
                            minIONdict[minion]["temp_history"]["heatsinktemp"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,heatsinktemp))
                            minIONdict[minion]["pore_history"]["strand"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,strand))
                            minIONdict[minion]["pore_history"]["percent"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,percent))
                            minIONdict[minion]["pore_history"]["single"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,single_pore))
                            minIONdict[minion]["pore_history"]["meanratio_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, meanratio))
                            minIONdict[minion]["pore_history"]["openpore_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, openpore))
                            minIONdict[minion]["pore_history"]["instrand_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, instrand))
                            if "simplesummary" in minIONdict[minion]:
                                if "details" not in minIONdict[minion]["pore_history"].keys():
                                    minIONdict[minion]["pore_history"]["details"]=dict()
                                for state in minIONdict[minion]["simplesummary"].keys():
                                    if state not in minIONdict[minion]["pore_history"]["details"].keys():
                                        minIONdict[minion]["pore_history"]["details"][state]=[]
                                    minIONdict[minion]["pore_history"]["details"][state].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,minIONdict[minion]["simplesummary"][state]))
                        else:
                            minIONdict[minion]["yield_history"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,yieldval))
                            minIONdict[minion]["temp_history"]["asictemp"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,asictemp))
                            minIONdict[minion]["temp_history"]["heatsinktemp"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,heatsinktemp))
                            minIONdict[minion]["temp_history"]["voltage"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,voltage_value))
                            minIONdict[minion]["pore_history"]["strand"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,strand))
                            minIONdict[minion]["pore_history"]["percent"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,percent))
                            minIONdict[minion]["pore_history"]["single"].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,single_pore))
                            minIONdict[minion]["pore_history"]["meanratio_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, meanratio))
                            minIONdict[minion]["pore_history"]["openpore_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, openpore))
                            minIONdict[minion]["pore_history"]["instrand_history"].append(
                                (minIONdict[minion]["detailsdata"]["timestamp"] * 1000, instrand))
                            if "simplesummary" in minIONdict[minion]:
                                if "details" not in minIONdict[minion]["pore_history"].keys():
                                    minIONdict[minion]["pore_history"]["details"]=dict()
                                for state in minIONdict[minion]["simplesummary"].keys():
                                    if state not in minIONdict[minion]["pore_history"]["details"].keys():
                                        minIONdict[minion]["pore_history"]["details"][state]=[]
                                    minIONdict[minion]["pore_history"]["details"][state].append((minIONdict[minion]["detailsdata"]["timestamp"]*1000,minIONdict[minion]["simplesummary"][state]))
                else:
                    minIONdict[minion]["yield_history"]=[]
                    minIONdict[minion]["temp_history"]["asictemp"]=[]
                    minIONdict[minion]["temp_history"]["heatsinktemp"]=[]
                    minIONdict[minion]["temp_history"]["voltage"]=[]
                    minIONdict[minion]["pore_history"]["strand"]=[]
                    minIONdict[minion]["pore_history"]["percent"]=[]
                    minIONdict[minion]["pore_history"]["single"]=[]
                    minIONdict[minion]["pore_history"]["openpore_history"] = []
                    minIONdict[minion]["pore_history"]["instrand_history"] = []
                    minIONdict[minion]["pore_history"]["meanratio_history"] = []
                    if "details" in minIONdict[minion]["pore_history"]:
                        minIONdict[minion]["pore_history"].pop("details")

def logitem(channel,minion,state):
    if minion not in channel_data:
        channel_data[minion]=dict()
        for i in chanlookup_type(minion):
            channel_data[minion][i]=dict()
    channel_data[minion][channel]["state"]=state


def process_channel_information(minIONdict,statedict, statesummarydict):
    """
    This function maintains the state of the channels in a simple to read format
    """
    colourlookup={}
    for minion in minIONdict:
        if "detailsdata" in minIONdict[minion]:
            if "channelstuff" in minIONdict[minion]:
                for item in minIONdict[minion]["channelstuff"]:
                    if 'style' in minIONdict[minion]["channelstuff"][item]:
                        colourlookup[minIONdict[minion]["channelstuff"][item]['style']['label']]=minIONdict[minion]["channelstuff"][item]['style']['colour']
                if "cust_chan_dict" not in minIONdict[minion]:
                    minIONdict[minion]["cust_chan_dict"]=dict()
                    for i in chanlookup_type(minion):
                        minIONdict[minion]["cust_chan_dict"][i]=dict()
                for channel in minIONdict[minion]["detailsdata"]["channel_info"]["channels"]:
                    if minion not in statedict:
                        statedict[minion]=dict()
                    if "state_group" in channel:
                        if channel["state"] in colourlookup:
                            statedict[minion].update({channel["name"]:{'state':channel["state"],"state_group":channel["state_group"],"colour":colourlookup[channel["state"]]}})
                            minIONdict[minion]["cust_chan_dict"][int(channel["name"])]=channel["state"]
                        elif channel["state_group"] in colourlookup:
                            statedict[minion].update({channel["name"]:{'state':channel["state"],"state_group":channel["state_group"],"colour":colourlookup[channel["state_group"]]}})
                            minIONdict[minion]["cust_chan_dict"][int(channel["name"])]=channel["state"]
                        else:
                            pass
                statesummarydict[minion]=dict()
                statesummarydict[minion]=get_state_summary(minion,channel_data)
                if "simplechanstats" not in minIONdict[minion]:
                    minIONdict[minion]["simplechanstats"]={}
                try:
                    minIONdict[minion]["simplesummary"]=statesummarydict[minion]
                except Exception as err:
                    print ("line 1081",err)
                    pass

def get_state_summary(minion,channel_data):
    state_dict = dict()
    for key, value in channel_data[minion].items():
        try:
            if value["state"] in state_dict:
                state_dict[value["state"]] += 1
            else:
                state_dict[value["state"]] = 1
        except Exception as err:
            print ("line 956", err)
            pass
    return state_dict


class DummyClient(WebSocketClient):

    def __init__(self, *args,**kwargs):
        super(DummyClient, self).__init__(*args,**kwargs)
        self.detailsdict=dict()
        self.daemon=True

    def init_minion(self,minion,minIONdict):
        """ ejecteject has three states:
            0=initisalised or sequencing finished
            1=sequencing happening
            2=run finish detected."""
        self.minion=minion
        self.ejecteject=0
        self.runname=""
        self.minIONdict=minIONdict


    def opened(self):
        self.send(json.dumps({'engine_states':'1','channel_states':'1','channel_info':'1'}))

    def closed(self, code, reason="client disconnected for some reason"):
        print ("socket",self.sock.fileno())
        print ("Closed down", code, reason)

    def received_message(self, m):
        if not m.is_binary:
            json_object = json.loads(str(m))
            for element in json_object:
                if element == "statistics" and json_object[element] != "null":
                    for element2 in json_object[element]:
                        if element2 != "read_statistics" and element2 != "channel_info" and json_object[element][element2] != "null":# and element2 != "read_statistics" and element2 != "completed_samples":
                            if element not in self.detailsdict:
                                self.detailsdict[element]=dict()
                            self.detailsdict[element][element2]=json_object[element][element2]
                elif element == "engine_states" and json_object[element] != "null":
                    for element2 in json_object[element]:
                        if json_object[element][element2] != "null":
                            if element not in self.detailsdict:
                                self.detailsdict[element]=dict()
                            if json_object[element][element2] is not dict:
                                self.detailsdict[element][element2]=json_object[element][element2]
                    if "status" in json_object[element]:
                        if json_object[element]["status"]== "finishing":
                            self.ejecteject=2
                        if json_object[element]["status"]== "starting":
                            print ("Looks as though a run is starting!!!!")
                            if self.ejecteject==0:
                                if self.minion in self.minIONdict and "yield_history" in self.minIONdict[self.minion]:
                                    self.minIONdict[self.minion]["yield_history"]=[]
                                    self.minIONdict[self.minion]["temp_history"]["asictemp"]=[]
                                    self.minIONdict[self.minion]["temp_history"]["heatsinktemp"]=[]
                                    self.minIONdict[self.minion]["temp_history"]["voltage"]=[]
                                    self.minIONdict[self.minion]["pore_history"]["strand"]=[]
                                    self.minIONdict[self.minion]["pore_history"]["percent"]=[]
                                    self.minIONdict[self.minion]["pore_history"]["single"]=[]
                                    if "details" in self.minIONdict[self.minion]["pore_history"]:
                                        self.minIONdict[self.minion]["pore_history"].pop("details")
                        if json_object[element]["status"]=="processing" and self.ejecteject==0:
                            print ("we're off - resetting history")
                            if self.minion in self.minIONdict and "yield_history" in self.minIONdict[self.minion]:
                                self.minIONdict[self.minion]["yield_history"]=[]
                                self.minIONdict[self.minion]["temp_history"]["asictemp"]=[]
                                self.minIONdict[self.minion]["temp_history"]["heatsinktemp"]=[]
                                self.minIONdict[self.minion]["temp_history"]["voltage"]=[]
                                self.minIONdict[self.minion]["pore_history"]["strand"]=[]
                                self.minIONdict[self.minion]["pore_history"]["percent"]=[]
                                self.minIONdict[self.minion]["pore_history"]["single"]=[]
                                if "details" in self.minIONdict[self.minion]["pore_history"]:
                                    self.minIONdict[self.minion]["pore_history"].pop("details")
                            self.ejecteject=1
                elif element == "channel_info" and json_object[element] != "null":
                    if "channels" in json_object[element].keys():
                        for channel in json_object[element]["channels"]:
                            state = "unknown"
                            if "state" in channel.keys():
                                state = channel["state"]
                            logitem(int(channel["name"]),self.minion,state)
                    for element2 in json_object[element]:
                        if json_object[element][element2] != "null":
                            if element not in self.detailsdict:
                                self.detailsdict[element]=dict()
                            if json_object[element][element2] is not dict:
                                self.detailsdict[element][element2]=json_object[element][element2]
                else:
                    if json_object[element] != "null":
                        self.detailsdict[element]=json_object[element]
            if self.ejecteject==2 and len(self.runname)>0:
                save_data()
                try:
                    if self.minion in self.minIONdict and "yield_history" in self.minIONdict[self.minion]:
                        self.minIONdict[self.minion]["yield_history"]=[]
                        self.minIONdict[self.minion]["temp_history"]["asictemp"]=[]
                        self.minIONdict[self.minion]["temp_history"]["heatsinktemp"]=[]
                        self.minIONdict[self.minion]["temp_history"]["voltage"]=[]
                        self.minIONdict[self.minion]["pore_history"]["strand"]=[]
                        self.minIONdict[self.minion]["pore_history"]["percent"]=[]
                        self.minIONdict[self.minion]["pore_history"]["single"]=[]
                        if "details" in self.minIONdict[self.minion]["pore_history"]:
                            self.minIONdict[self.minion]["pore_history"].pop("details")
                    self.ejecteject=0
                except Exception as err:
                    print ("Problem",err)
                    print ("Probably already popped")


###API shizzle

class MinControlAPI:
    """
    Notes for requests
    header = {'Authorization':'Token e45c142b457121278f9b67d713285a7e10382b36', 'Content-Type':'application/json'}
    r=requests.post(args.full_host+'api/v1/runs/', headers=header, json={"run_name": "20170612_1630_matt", "run_id": "hj78yy9o-217e-4335-9451-66a7288a9dd5", "barcode": "barcode10"})
    r.text #returns result
    """

    def __init__(self, minion_name, args, statedict, summarystatedict, minIONdict, header):

        self.args = args
        self.minIONdict=minIONdict
        self.header = header

        self.runidlink = ""
        self.status_summary=dict()
        self.minstatexist=False
        self.current_run_id=""
        self.computer=""
        self.runid = 0
        self.flowcelllink = ""
        self.statedict = statedict
        self.summarystatedict = summarystatedict

        self.minotourapi = MinotourAPINew(self.args.full_host, self.header)

        minion = self.minotourapi.get_minion_by_name(minion_name)

        if not minion:

            minion = self.minotourapi.create_minion(minion_name)

        self.minion = minion
        self.minionidlink=self.identify_minion(self.minion["url"])


    # def header (self):
    #     return ({'Authorization': 'Token ' + self.args.api_key, 'Content-Type': 'application/json'})

    def check_jobs (self,minION):
        if self.args.GUI:
            self.args.minKNOWmessage = "Checking Jobs"
        #print ("checking jobs")
        r = requests.get(self.minion["url"] + 'control/', headers=self.header)
        #print (r.text)
        jobs = json.loads(r.text)
        for job in jobs:
            #print (job['job'])

            if job["job"] == "testmessage":
                send_message_port("minoTour is checking communication status with " + str(self.minion['name']) + ".",
                                  self.args.ip,self.minIONdict[self.minion['name']]["port"])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print (r.text)

            if job["job"] == "custommessage":

                print(self.minIONdict)

                send_message_port(
                    str(job["custom"]),
                    self.args.ip,
                    self.minIONdict[self.minion['name']]["port"]
                )

                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print (r.text)

            if job["job"] == "stopminion":
                stoprun(self.args.ip,self.minIONdict[self.minion['name']]["port"])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)

            if job["job"] == "rename":
                renamerun(job["custom"], self.args.ip,self.minIONdict[self.minion['name']]["port"])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)

            if job["job"] == "nameflowcell":
                renameflowcell(job["custom"], self.args.ip,self.minIONdict[self.minion['name']]["port"])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)



            if job["job"] == "startminion":
                startrun(job["custom"],self.args.ip, self.minIONdict[self.minion['name']]["port"])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)

            if job["job"] == "initialiseminion":
                # print "Trying to initialise minION"
                # result = helper.initialiseminion()
                startstop('start', self.minion['name'])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)
                # execute_command_as_string(commands('initialiseminion'), self.args.ip,self.minIONdict[message["minion"]]["port"])

            if job["job"] == "shutdownminion":
                # First we need to stop the minion if it is running...
                # stoprun(self.minIONdict[message["minion"]]["port"])
                # execute_command_as_string(commands('shutdownminion'), self.args.ip,self.minIONdict[message["minion"]]["port"])
                startstop('stop', self.minion['name'])
                r = requests.post(self.minion["url"] + 'control/' + str(job["id"]) + '/', headers=self.header)
                #print(r.text)

    def create_run(self, runid):

        run = self.minotourapi.get_run_by_runid(runid)

        if not run:

            #
            # get or create a flowcell
            #
            flowcell = self.minotourapi.get_flowcell_by_name(self.status_summary['flow_cell_id'])

            if not flowcell:

                flowcell = self.minotourapi.create_flowcell(self.status_summary['flow_cell_id'])

            is_barcoded = False  # TODO do we known this info at this moment?

            has_fastq = True  # TODO do we known this info at this moment?

            if self.minion:

                print("minion exists")

            else:

                print("minion does not exist")

            createrun = self.minotourapi.create_run(self.status_summary['run_name'], runid, is_barcoded, has_fastq, flowcell, self.minion)

            # createrun = requests.post(self.args.full_host+'api/v1/runs/', headers=self.header, json={"run_name": self.status_summary['run_name'], "run_id": runid, "barcode": barcoded, "is_barcoded":is_barcoded, "minION":self.minion["url"]})

            if not createrun:

                print("Houston - we have a problem!")

            else:

                self.runidlink = createrun["url"]
                self.runid = createrun["id"]  # TODO is it id or runid?

        else:

            self.runidlink = run["url"]

        self.update_minion_run_stats()

    def update_minion_run_stats(self):

        payload = {
            "minION": self.minion["url"],
            "minKNOW_current_script": self.status_summary['current_script'],
            "minKNOW_sample_name": self.status_summary['sample_name'],
            "minKNOW_exp_script_purpose": self.status_summary['exp_script_purpose'],
            "minKNOW_flow_cell_id": self.status_summary['flow_cell_id'],
            "minKNOW_run_name": self.status_summary['run_name'],
            "run_id": self.runidlink,
            "minKNOW_version": self.status_summary['minKNOW_version'],
            "minKNOW_hash_run_id": self.status_summary['hash_run_id'],
            "minKNOW_script_run_id": self.status_summary['script_run_id'],
            "minKNOW_real_sample_rate": self.status_summary['real_sample_rate'],
            "minKNOW_asic_id": self.status_summary['asic_id'],
            "minKNOW_start_time": self.status_summary['start_time'],
            "minKNOW_colours_string": self.status_summary['colour_data'],
            "minKNOW_computer": self.computer,
        }

        requests.post(
            self.runidlink + 'rundetails/',
            headers=self.header,
            json=payload
        )

    def update_message(self,messagestring):
        payload = {"minION":self.minion["url"],
                   "run_id":self.runidlink,
                   "minKNOW_message":messagestring["message"],
                   "minKNOW_identifier": messagestring["identifier"],
                   "minKNOW_severity": messagestring["severity"],
                   "minKNOW_message_timestamp": messagestring["timestamp"],
        }
        #print (self.minion["url"] + 'messages/')
        #print (payload)
        createminIONStatus = requests.post(self.minion["url"] + 'messages/', headers=self.header,
                                           json=payload)


    def identify_minion(self,minION):
        #print ("args full host {}".format(self.args.full_host))
        r = requests.get(self.args.full_host + 'api/v1/minions', headers=self.header)
        minionidlink = ""
        for minion in json.loads(r.text):
            if minion["minION_name"] == minION:
                minionidlink = minion["url"]
        return(minionidlink)

    def update_minion_current_stats (self,livedata,detailsdata,simplesummary,channelstuff):
        try:
            if self.args.GUI:
                self.args.minKNOWmessage = livedata['status']['result']
                self.args.minKNOWreadsseen = livedata["get_seq_metrics"]["result"]["selected_completed_count"]
            self.status_summary['colour_data']= json.dumps(channelstuff)
            self.status_summary['start_time'] = datetime.datetime.fromtimestamp(
        int(livedata['get_engine_states']['result']['daq_start_time'])
    ).strftime('%Y-%m-%d %H:%M:%S')
            self.status_summary['status'] = livedata['status']['result']
            self.status_summary['current_script'] = livedata['current_script']['result']
            self.status_summary['sample_name']=livedata['get_engine_states']['result']['sample_id']
            self.status_summary['minKNOW_version']=livedata['get_engine_states']['result']['version_string']
            self.status_summary['flow_cell_id'] = livedata['get_engine_states']['result']['flow_cell_id']
            self.status_summary['run_name'] = livedata['get_engine_states']['result']['data_set']
            self.status_summary['hash_run_id'] = livedata['get_engine_states']['result']['hash_run_id']
            self.status_summary['exp_script_purpose'] = livedata['get_engine_states']['result']['exp_script_purpose']
            self.status_summary['script_run_id'] = livedata['get_engine_states']['result']['script_run_id']
            self.status_summary['real_sample_rate'] = livedata['get_engine_states']['result']['real_sample_rate']
            self.status_summary['asic_id'] = livedata['get_engine_states']['result']['asic_id']
            self.status_summary['bytes_capacity'] = livedata['disk_space']['result'][0]['bytes_capacity']
            self.status_summary['bytes_available'] = livedata['disk_space']['result'][0]['bytes_available']
            self.status_summary['bytes_when_alert_issued'] = livedata['disk_space']['result'][0]['bytes_when_alert_issued']
            self.status_summary['recommend_alert'] = livedata['disk_space']['result'][0]['recommend_alert']
            self.status_summary['histogram']=str(livedata["get_seq_metrics"]["result"]["read_event_count_weighted_hist"])
            self.status_summary['bin_width']=livedata["get_seq_metrics"]["result"]["read_event_count_weighted_hist_bin_width"]
            self.status_summary['read_count'] = livedata["get_seq_metrics"]["result"]["selected_completed_count"]

            payload = {"minION": self.minion["url"],
                    "minKNOW_status": self.status_summary['status'],
                    "minKNOW_current_script": self.status_summary['current_script'],
                    "minKNOW_sample_name": self.status_summary['sample_name'],
                    "minKNOW_exp_script_purpose": self.status_summary['exp_script_purpose'],
                    "minKNOW_flow_cell_id": self.status_summary['flow_cell_id'],
                    "minKNOW_run_name": self.status_summary['run_name'],
                    "minKNOW_hash_run_id": self.status_summary['hash_run_id'],
                    "minKNOW_script_run_id": self.status_summary['script_run_id'],
                    "minKNOW_real_sample_rate": self.status_summary['real_sample_rate'],
                    "minKNOW_asic_id": self.status_summary['asic_id'],
                    "minKNOW_total_drive_space": self.status_summary['bytes_capacity'],
                    "minKNOW_disk_space_till_shutdown": self.status_summary['bytes_when_alert_issued'],
                    "minKNOW_disk_available": self.status_summary['bytes_available'],
                    "minKNOW_warnings": self.status_summary['recommend_alert'],
                }
            if self.minstatexist is False:
                r = requests.get(self.minion["url"] + 'status/', headers=self.header)
                if r.status_code == 200:
                    self.minstatexist = True
                else:
                    createminIONStatus = requests.post(self.minion["url"] + 'status/', headers=self.header,
                                                      json=payload)
            else:
                createminIONStatus = requests.put(self.minion["url"] + 'status/', headers=self.header,
                                             json=payload)
            if self.status_summary['status'] == 'processing':

                print(">>> inside self.status_summary['status'] == 'processing'")
                if self.current_run_id != self.status_summary['hash_run_id']:

                    print(">>> inside self.current_run_id != self.status_summary['hash_run_id']")

                    print(">>> before create_run")
                    self.create_run(self.status_summary['hash_run_id'])

                    print(">>> before current_run_id")
                    self.current_run_id = self.status_summary['hash_run_id']

                    print(">>> after current_run_id")

                self.update_minion_stats(livedata,detailsdata,simplesummary)
        except Exception as err:
            print ("Problem",err)

    def update_minion_stats (self,livedata,detailsdata,simplesummary):
        if "yield_res" in livedata:
            if "result" in livedata["yield_res"]:
                try:
                    asictemp = float(detailsdata["engine_states"]["minion_asic_temperature"])
                except:
                    asictemp = 0
                try:
                    heatsinktemp = float(
                        detailsdata["engine_states"]["minion_heatsink_temperature"])
                except:
                    heatsinktemp = 0
                if livedata["bias_voltage_gain"]["result"] != "null":
                    # print "!!!!!!!!STUFF!!!!!", self.minIONdict[minion]["livedata"]["bias_voltage_gain"]["result"]
                    biasvoltage = int(livedata["bias_voltage_gain"]["result"])
                else:
                    biasvoltage = 0
                if livedata["biasvoltageget"]["result"] != "null":
                    voltage_val = int(livedata["biasvoltageget"]["result"]["bias_voltage"])
                else:
                    voltage_val = 0
                voltage_value = voltage_val * biasvoltage
                # print "volts:",voltage_value
                try:
                    if livedata["yield_res"]["result"] != "null":
                        yieldval = int(livedata["yield_res"]["result"])
                        if self.args.GUI:
                            self.args.minKNOWeventsseen = int(livedata["yield_res"]["result"])

                    else:
                        yieldval = detailsdata["statistics"]["read_event_count"]
                        if self.args.GUI:
                            self.args.minKNOWeventsseen = detailsdata["statistics"]["read_event_count"]

                except:
                    yieldval = 0
                try:
                    meanratio = detailsdata["meanratio"]
                    openpore = detailsdata["openpore"]
                    instrand = detailsdata["instrand"]
                except:
                    meanratio = 0
                    openpore = 0
                    instrand = 0

                payload = {"minION": self.minion["url"],
                   "run_id": self.runidlink,
                   "sample_time": str(datetime.datetime.now()),
                   "event_yield": int(livedata['get_seq_metrics']['result']['read_event_count']),
                   "asic_temp": asictemp,
                   "heat_sink_temp": heatsinktemp,
                   "voltage_value": voltage_value,
                   "mean_ratio": meanratio,
                   "open_pore":openpore,
                   "in_strand":instrand,
                   "minKNOW_histogram_values": self.status_summary['histogram'],
                   "minKNOW_histogram_bin_width": self.status_summary['bin_width'],
                    "minKNOW_read_count": self.status_summary['read_count']
                   }

                for category in simplesummary:
                    #print (category)
                    payload[str(category)]=simplesummary[category]
                #print (payload)
                createminionstat = requests.post(self.runidlink + 'runstats/', headers=self.header, json=payload)
                #print (createminionstat.text)
                #print(createminionstat.status_code)


    def update_minion_status (self,minION,computer,status):
        self.computer = computer
        #print (self.minion["url"])
        #Second get id of status we are looking for:
        statusidlink=""
        r = requests.get(self.args.full_host + 'api/v1/events', headers=self.header)
        #print (r.text)
        for info in json.loads(r.text):
            if info["name"] == status:
                #print(info)
                statusidlink= info["url"]

        payload = {"computer_name": computer, "datetime": str(datetime.datetime.now()), "event": str(urlparse(statusidlink).path),"minION": str(self.minion["url"])}

        updatestatus = requests.post(
            self.minion["url"] + "events/",
            headers=self.header,
            json=payload
        )
        #print (updatestatus.text)

    def check_scripts(self,minION):
        r=requests.get(self.minion["url"] + 'scripts/', headers = self.header)
        self.scripts=r.text

    def update_script(self,minION,script):
        #print ("Updating script registry for {} with {}".format(minION,script))
        #minionidlink=self.identify_minion(minION)
        #print (script['name'])
        #print (script["tags"])
        #Check if we have seen the script before
        if script["name"] not in self.scripts:
            #so script is not present.
            payload=dict()
            payload["minION"]=self.minion["url"]
            payload["identifier"]= script["identifier"]
            payload["name"]= script["name"]
            if "experiment type" in script["tags"].keys():
                payload["experiment_type"] = script["tags"]["experiment type"]
            if "flow cell" in script["tags"].keys():
                payload["flow_cell"]=script["tags"]["flow cell"]
            if "base calling" in script["tags"].keys():
                payload["base_calling"]=script["tags"]["base calling"]
            if "kit" in script["tags"].keys():
                payload["kit"]=script["tags"]["kit"]
            #print ("PAYLOAD", payload)
            updatestate = requests.post(self.minion["url"] + 'scripts/', json=payload, headers=self.header)
        else:
            #so script exists, but is the identifier the same or different?
            #print ("so script exists, but is the identifier the same or different?")
            #print (script["identifier"])
            #print (self.scripts)
            if str(script["identifier"]) not in self.scripts:
                #print ("path has changed")
                for element in json.loads(self.scripts):
                    if script["name"] in str(element):
                        #print (element)
                        payload = dict()
                        payload["minION"] = self.minion["url"]
                        payload["identifier"] = script["identifier"]
                        payload["name"] = script["name"]
                        if "experiment type" in script["tags"].keys():
                            payload["experiment_type"] = script["tags"]["experiment type"]
                        if "flow cell" in script["tags"].keys():
                            payload["flow_cell"] = script["tags"]["flow cell"]
                        if "base calling" in script["tags"].keys():
                            payload["base_calling"] = script["tags"]["base calling"]
                        if "kit" in script["tags"].keys():
                            payload["kit"] = script["tags"]["kit"]
                        update = requests.put(self.minion["url"] + "scripts/" + str(element["id"]) + "/", json=payload, headers=self.header )
                        #print (update.text)
            else:
                pass


class HelpTheMinion(WebSocketClient):
    def __init__(self, minswip, args, header):
        WebSocketClient.__init__(self, minswip)
        self.args = args
        self.minIONdict = dict()
        self.minIONdict_test = dict()
        self.minIONclassdict = dict()
        self.statedict = dict()
        self.summarystatedict = dict()
        self.mcrunning = False
        self.header = header
        #t = threading.Thread(target=self.process_minion_test())
        #t.daemon = True
        #t.start()

    def opened(self):
        if self.args.GUI:
            self.args.minKNOWmessage = "Connected to MinKNOW"
        print ("Connected to Master MinION Controller!")

    def initialiseminion():
        if self.args.GUI:
            self.args.minKNOWmessage = "Initialising MinION"
        #result = self.send(json.dumps({"params": "null", "id": 5, "method": "initialization_status"}))
        #print (result)

    def determinetype(self,minION):
        """
        :param minION:
        :return: devicetype,deviceid,portstring
        """
        devicetype = "unknown"
        deviceid = "unknown"
        if minION[1] == "M":
            devicetype = "MinION"
            deviceid=minION[1:8]
        elif minION[1] == "G":
            devicetype = "GridION"
            deviceid = minION[1:8]
        elif minION[1] =="1" or minION[1]=="2":
            devicetype = "PromethION"
            #print (minION[1:9])
            promvals = minION[1:9].split('-')
            if len(promvals[1])==2:
                print (minION[1:8])
                deviceid = minION[1:8]
            else:
                print (minION[1:10])
                deviceid = minION[1:10]
        print (devicetype)
        return devicetype,deviceid  #,portstring

    def parse_ports(self,ports,minIONname):
        """
        :param ports: string to parse representing the ports
        :return: list of ports
        """
        portset = list()
        rawports = list(map(lambda x: x, filter(lambda x: x, map(ord, ports))))
        chomplen = len(minIONname)+4
        print (rawports)
        rawports = rawports[chomplen:-2]
        rawvalues = list()
        offset = list()
        for index,port in enumerate(rawports):
            if index%3==0:
                try:
                    rawvalues.append(port)
                    offset.append(rawports[(index+1)])
                except:
                    print ("fail")
        #print (rawports)
        #print (rawvalue)
        #print (offset)

        for index,rawvalue in enumerate(rawvalues):
            try:
                off = offset[index]
                if off == 62:
                    correction = 192
                elif off == 63:
                    correction = 64
                else:
                    print("These are not the ports you are looking for.")
                    correction = 0
                    #sys.exit()
                port = rawvalue - correction + 8000
                print (port)
                portset.append(port)
            except:
                print ("fail2")
        return portset

    def received_message(self, m):
        for thing in ''.join(map(chr,map(ord,(m.data).decode('latin-1')))).split('\n'):
            if len(thing) > 5 and "2L" not in thing and "2n" not in thing:
                devicetype,deviceid=self.determinetype(thing)
                if len(thing) > 5 and (thing[1] == "M" or thing[1] == "G" or thing[1] =="1" or thing[1] =="2"):
                    if deviceid not in self.minIONdict:
                        #print ("INITIALISING self.minIONdict", deviceid)
                        self.minIONdict[deviceid]=dict()
                        self.minIONdict[deviceid]["APIHelp"] = MinControlAPI(deviceid, self.args, self.statedict, self.summarystatedict, self.minIONdict, self.header)
                        self.minIONdict[deviceid]["APIHelp"].update_minion_status(deviceid,'UNKNOWN','connected')
                    # minIONports = list(map(lambda x:x,filter(lambda x:x,map(ord,thing))))
                    # print (minIONports)
                    # minIONports = list(map(lambda x: x, filter(lambda x: x > 120, map(ord, thing))))
                    # print (minIONports)
                    # minIONports = list(map(lambda x:x-192+8000,filter(lambda x:x>120,map(ord,thing))))
                    # print (minIONports)
                    minIONports = self.parse_ports(thing,deviceid)
                    print (minIONports)

                    # time.sleep(1)
                    if len(minIONports) > 3:
                        # if min(minIONports) != minIONports[0]:
                        #     print ("These are not the ports you are looking for.")
                        #     minIONports = list(
                        #         map(lambda x: x - 192 + 8000 + 128, filter(lambda x: x > 120, map(ord, thing))))
                        #     print (minIONports)
                        try:
                            self.minIONdict[deviceid]["state"]="active"
                            port = minIONports[0]
                            ws_longpoll_port = minIONports[1]
                            ws_event_sampler_port = minIONports[2]
                            ws_raw_data_sampler_port = minIONports[3]
                            self.minIONdict[deviceid]["port"]=port
                            self.minIONdict[deviceid]["ws_longpoll_port"]=ws_longpoll_port
                            self.minIONdict[deviceid]["ws_event_sampler_port"]=ws_event_sampler_port
                            self.minIONdict[deviceid]["ws_raw_data_sampler_port"]=ws_raw_data_sampler_port
                            # print ("port is:", port)
                            results = execute_command_as_string(commands("machine_id"), self.args.ip,
                                                           self.minIONdict[deviceid]["port"])

                            self.minIONdict[deviceid]["APIHelp"].update_minion_status(deviceid, str(results["result"]), 'active')
                        except:
                            # print ("caught it")
                            minIONports = list(
                                map(lambda x: x - 192 + 8000 +128, filter(lambda x: x > 120, map(ord, thing))))
                            # print (minIONports)
                            self.minIONdict[deviceid]["state"] = "active"
                            port = minIONports[0]
                            ws_longpoll_port = minIONports[1]
                            ws_event_sampler_port = minIONports[2]
                            ws_raw_data_sampler_port = minIONports[3]
                            self.minIONdict[deviceid]["port"] = port
                            self.minIONdict[deviceid]["ws_longpoll_port"] = ws_longpoll_port
                            self.minIONdict[deviceid]["ws_event_sampler_port"] = ws_event_sampler_port
                            self.minIONdict[deviceid]["ws_raw_data_sampler_port"] = ws_raw_data_sampler_port
                            # print ("port is:", port)
                            results = execute_command_as_string(commands("machine_id"), self.args.ip,
                                                                self.minIONdict[deviceid]["port"])
                            # print ("should contain the computer name", results)
                            self.minIONdict[deviceid]["APIHelp"].update_minion_status(deviceid,str(results["result"]),'active')
                        # print (minIONports)
                    else:
                        self.minIONdict[deviceid]["state"]="inactive"
                        self.minIONdict[deviceid]["port"]=""
                        self.minIONdict[deviceid]["ws_longpoll_port"]=""
                        self.minIONdict[deviceid]["ws_event_sampler_port"]=""
                        self.minIONdict[deviceid]["ws_raw_data_sampler_port"]=""
                        self.minIONdict[deviceid]["APIHelp"].update_minion_status(deviceid, 'UNKNOWN', 'inactive')
        self.mcrunning=True

    def process_minion_test(self):
        while True:
            #print ("yo dawg")
            time.sleep(1)

    def process_minion(self):
        while True:
            while self.mcrunning:
                print ("Still Monitoring")
                #print (".")
                #print (self.mcrunning)
                #print (minIONdict)
                if self.minIONdict == self.minIONdict_test: ## The dictionary is unchanged since the last cycle
                    active = 0
                    inactive = 0
                    for minION in self.minIONdict:
                        if self.minIONdict[minION]["state"] == "active":
                            active += 1
                            #print ("minION is active")
                            if minION not in self.minIONclassdict:
                                #print ("adding minION to dict")
                                self.minIONclassdict[minION] = dict()
                            if "connected" not in self.minIONclassdict[minION]:
                                connectip = "ws://" + self.args.ip + ":" + str(self.minIONdict[minION]["ws_longpoll_port"]) + "/"
                                self.minIONclassdict[minION]["class"] = DummyClient(connectip)
                                connectip2 = "ws://" + self.args.ip + ":" + str(
                                    self.minIONdict[minION]["ws_longpoll_port"]) + "/user_messages"
                                self.minIONclassdict[minION]["class2"] = MessagesClient(connectip2,self.args)
                                #print ("initial handshake")
                            try:
                                if "connected" not in self.minIONclassdict[minION]:
                                    try:
                                        self.minIONclassdict[minION]["class"].connect()
                                        self.minIONclassdict[minION]["class2"].connect()
                                        self.minIONclassdict[minION]["class"].init_minion(minION,self.minIONdict)
                                        print ("Trying to init a messages client with {}".format(minION))
                                        self.minIONclassdict[minION]["class2"].init_minion(minION)
                                        results = execute_command_as_string(commands('get_analysis_configuration'), self.args.ip,
                                                                            self.minIONdict[minION]["port"])
                                        #print ("getting the good stuff")

                                        self.minIONdict[minION]["channelstuff"] = results["result"]["channel_states"]
                                        #print ("connection good")
                                    except Exception as err:
                                        print ("380 Connection failed", err)
                                    self.minIONclassdict[minION]["connected"] = "True"
                                ##Check if connction is new or old:
                                if minION in self.minIONdict_test:
                                    if self.minIONdict[minION]["state"] == self.minIONdict_test[minION]["state"]:
                                        try:  # To catch mysterious bugs
                                            livedata = dict()  # collect a dictionary of useful data - might change this
                                            for query in (
                                            'status', 'dataset', 'biasvoltageget', 'bias_voltage_gain', 'machine_id',
                                            'machine_name', 'sample_id', 'user_error', 'sequenced_res', 'yield_res',
                                            'current_script', 'disk_space', 'flow_cell_id',
                                            'get_seq_metrics', 'get_engine_states'):  # ,'getstaticdata','get_analysis_configuration'):
                                                results = execute_command_as_string(commands(query), self.args.ip,
                                                                                    self.minIONdict[minION]["port"])
                                                livedata[query] = results
                                                #print (query,results)
                                            self.minIONdict[minION]["livedata"] = livedata  # append results to data stream
                                            #print ("data appended")
                                            #sys.exit()
                                        except Exception as err:
                                            print ("line 1407", err)
                                            pass
                                        #print ("OK")
                                    else:
                                        if "scripts" in self.minIONdict[minION]:
                                            #print ("running get run scripts line 1110")
                                            if self.minIONdict[minION]["scripts"] != get_run_scripts(self.args.ip,self.minIONdict[minION]["port"]):
                                                #print("running get run scripts 1111")
                                                self.minIONdict[minION]["scripts"] = get_run_scripts(self.args.ip,self.minIONdict[minION]["port"])
                                                #print ("calling check scripts 1113")
                                                self.minIONdict[minION]["APIHelp"].check_scripts(minION)
                                                for script in self.minIONdict[minION]["scripts"]:
                                                    # print (script)
                                                    self.minIONdict[minION]["APIHelp"].update_script(minION, script)
                                        else:
                                            #print("running get run scripts 1114")
                                            self.minIONdict[minION]["scripts"] = get_run_scripts(self.args.ip,self.minIONdict[minION]["port"])
                                            #print("calling check scripts 1121")
                                            self.minIONdict[minION]["APIHelp"].check_scripts(minION)
                                            for script in self.minIONdict[minION]["scripts"]:
                                                # print (script)
                                                self.minIONdict[minION]["APIHelp"].update_script(minION, script)
                                            results = execute_command_as_string(commands('startmessagenew'), self.args.ip,
                                                                                self.minIONdict[minION]["port"])
                                            try:  # To catch mysterious bugs
                                                livedata = dict()  # collect a dictionary of useful data - might change this
                                                for query in (
                                                'status', 'dataset', 'biasvoltageget', 'bias_voltage_gain', 'machine_id',
                                                'machine_name', 'sample_id', 'user_error', 'sequenced_res', 'sequenced_res',
                                                'yield_res', 'current_script', 'disk_space', 'flow_cell_id',
                                                'get_seq_metrics'):  # ,'getstaticdata','get_analysis_configuration'):
                                                    # print query
                                                    results = execute_command_as_string(commands(query), self.args.ip,
                                                                                        self.minIONdict[minION]["port"])
                                                    livedata[query] = results
                                                self.minIONdict[minION]["livedata"] = livedata  # append results to data stream
                                            except Exception as err:
                                                print ("line 1438", err)
                                                pass
                            except Exception as err:
                                print ("line 1443", err)
                                print ("Connection Error")
                            else:
                                #print ("here")
                                if "scripts" not in self.minIONdict[minION]:
                                    #print ("running get run scripts 1147")
                                    self.minIONdict[minION]["scripts"] = get_run_scripts(self.args.ip,self.minIONdict[minION]["port"])
                                    #print("calling check scripts 1161")
                                    self.minIONdict[minION]["APIHelp"].check_scripts(minION)
                                    for script in self.minIONdict[minION]["scripts"]:
                                        # print (script)
                                        self.minIONdict[minION]["APIHelp"].update_script(minION, script)
                                    #print (get_run_scripts(self.args.ip,self.minIONdict[minION]["port"]))
                                    #APIHelp = MinControlAPI()
                                    #print("calling check scripts 1168")
                                    #self.minIONdict[minION]["APIHelp"].check_scripts(minION)
                                    for script in self.minIONdict[minION]["scripts"]:
                                        #print (script)
                                        self.minIONdict[minION]["APIHelp"].update_script(minION,script)
                                    #time.sleep(5)
                                    #sys.exit()
                        else:
                            inactive += 1
                        process_tracked_yield(self.minIONdict)
                        process_channel_information(self.minIONdict, self.statedict, self.summarystatedict)
                        self.minIONdict[minION]["APIHelp"].update_minion_current_stats(self.minIONdict[minION]['livedata'],self.minIONdict[minION]['detailsdata'],self.minIONdict[minION]["simplesummary"],self.minIONdict[minION]["channelstuff"])
                        self.minIONdict[minION]["APIHelp"].check_jobs(minION)
                else:
                    #print ("Dictionary has changed")
                    for minION in self.minIONdict:  # if yes then we check each minION to see what is happening
                        if "state" in self.minIONdict[minION].keys():
                            if self.minIONdict[minION]["state"] == "active":  # we have a sequencing minION - what is it doing?
                                #APIHelp =
                                #print (minION, "ACTIVE")
                                try:
                                    if minION not in self.minIONclassdict:
                                        connectip = "ws://" + self.args.ip + ":" + str(
                                            self.minIONdict[minION]["ws_longpoll_port"]) + "/"  # Connect to the self.minIONdict
                                        connectip2 = "ws://" + self.args.ip + ":" + str(self.minIONdict[minION][
                                                                                       "ws_longpoll_port"]) + "/user_messages"  # Connect to the self.minIONdict
                                        self.minIONclassdict[minION] = dict()  # Add minION to the dictionary
                                    if "connected" not in self.minIONclassdict[minION]:
                                        #print ("Connecting to this minION", minION)
                                        self.minIONclassdict[minION]["class"] = DummyClient(connectip)
                                        self.minIONclassdict[minION]["class2"] = MessagesClient(connectip2,self.args,self.minIONdict)
                                    if "connected" not in self.minIONclassdict[minION]:
                                        try:
                                            self.minIONclassdict[minION]["class"].connect()
                                            self.minIONclassdict[minION]["class2"].connect()
                                        except Exception as err:
                                            print ("480 Connection failed", err)
                                        try:
                                            #print ("GETTING THE GOOD STUFF")
                                            self.minIONclassdict[minION]["class"].init_minion(minION,self.minIONdict)
                                            print ("Trying to set the minion Again.... {}".format(minION))
                                            self.minIONclassdict[minION]["class2"].init_minion(minION)
                                            results = execute_command_as_string(commands('get_analysis_configuration'), self.args.ip,
                                                                                self.minIONdict[minION]["port"])
                                            self.minIONdict[minION]["channelstuff"] = results["result"]["channel_states"]
                                        except Exception as err:
                                            print ("Connection failed", err)
                                        self.minIONclassdict[minION]["connected"] = "True"
                                    livedata = dict()  # collect a dictionary of useful data - might change this
                                    #print ("running get run scripts 1199")
                                    self.minIONdict[minION]["scripts"] = get_run_scripts(self.args.ip,self.minIONdict[minION]["port"])
                                    #print("calling check scripts 1216")
                                    self.minIONdict[minION]["APIHelp"].check_scripts(minION)
                                    for script in self.minIONdict[minION]["scripts"]:
                                        # print (script)
                                        self.minIONdict[minION]["APIHelp"].update_script(minION, script)
                                    for query in (
                                            'status', 'dataset', 'biasvoltageget', 'bias_voltage_gain', 'machine_id',
                                            'machine_name',
                                            'sample_id', 'user_error', 'sequenced_res', 'sequenced_res', 'yield_res',
                                            'current_script',
                                            'disk_space', 'flow_cell_id', 'get_tracking_id',
                                            'get_seq_metrics', 'get_engine_states'):  # ,'getstaticdata','get_analysis_configuration'):
                                        results = execute_command_as_string(commands(query), self.args.ip, self.minIONdict[minION]["port"])
                                        livedata[query] = results
                                    results = execute_command_as_string(commands('get_statistics'), self.args.ip,
                                                                        self.minIONdict[minION]["port"])
                                    #print ("Got Stats")
                                    meanratio = list()
                                    openpore = list()
                                    instrand = list()
                                    datatofetch = ('seq_pore_level', 'seq_strand_delta', 'tba_pore_level', 'tba_event_delta')
                                    for item in results['result']['stats_for_channel_name']:
                                        if float(results['result']['stats_for_channel_name'][item]['seq_strand_delta']) != 0:
                                            meanratio.append((float(
                                                results['result']['stats_for_channel_name'][item]['seq_pore_level']) / float(
                                                results['result']['stats_for_channel_name'][item]['seq_strand_delta'])))
                                            openpore.append(
                                                float(results['result']['stats_for_channel_name'][item]['seq_pore_level']))
                                            instrand.append(
                                                float(results['result']['stats_for_channel_name'][item]['seq_strand_delta']))

                                    results2 = execute_command_as_string(commands('get_analysis_configuration'), self.args.ip,
                                                                         self.minIONdict[minION]["port"])
                                    self.minIONdict[minION]["channelstuff"] = results2["result"]["channel_states"]
                                    self.minIONdict[minION]["livedata"] = livedata  # append results to data stream
                                    #print (livedata)
                                    if "class" in self.minIONclassdict[minION]:
                                        for element in self.minIONclassdict[minION]["class"].detailsdict:
                                            if "detailsdata" not in self.minIONdict[minION]:
                                                print ("Creating detailsdata")
                                                self.minIONdict[minION]["detailsdata"] = dict()
                                            try:
                                                self.minIONdict[minION]["detailsdata"]["meanratio"] = np.mean(meanratio)
                                                self.minIONdict[minION]["detailsdata"]["openpore"] = np.mean(openpore)
                                                self.minIONdict[minION]["detailsdata"]["instrand"] = np.mean(instrand)
                                            except:
                                                self.minIONdict[minION]["detailsdata"]["meanratio"] = 0
                                                self.minIONdict[minION]["detailsdata"]["openpore"] = 0
                                                self.minIONdict[minION]["detailsdata"]["instrand"] = 0
                                            if element == "statistics" and self.minIONclassdict[minION]["class"].detailsdict[
                                                element] != "null":
                                                if element not in self.minIONdict[minION]["detailsdata"]:
                                                    self.minIONdict[minION]["detailsdata"][element] = dict()
                                                context = dict(list(self.minIONdict[minION]["detailsdata"][element].items()) + list(
                                                    self.minIONclassdict[minION]["class"].detailsdict[element].items()))
                                                self.minIONdict[minION]["detailsdata"][element] = context
                                            if element == "engine_states" and self.minIONclassdict[minION]["class"].detailsdict[
                                                element] != "null":
                                                for element2 in self.minIONclassdict[minION]["class"].detailsdict[element]:
                                                    if self.minIONclassdict[minION]["class"].detailsdict[element][
                                                        element2] != "null":
                                                        if element not in self.minIONdict[minION]["detailsdata"]:
                                                            self.minIONdict[minION]["detailsdata"][element] = dict()
                                                        self.minIONdict[minION]["detailsdata"][element][element2] = \
                                                        self.minIONclassdict[minION]["class"].detailsdict[element][element2]
                                            else:
                                                if self.minIONclassdict[minION]["class"].detailsdict[element] != "null":
                                                    self.minIONdict[minION]["detailsdata"][element] = \
                                                    self.minIONclassdict[minION]["class"].detailsdict[element]
                                except Exception as err:
                                    self.minIONclassdict.pop(minION)
                                    print ("line 1411 minCONTROL", err)
                                    print ("Connection Error")
                                process_tracked_yield(self.minIONdict)
                                process_channel_information(self.minIONdict, self.statedict, self.summarystatedict)
                                if "livedata" in self.minIONdict[minION].keys():
                                    try:
                                        self.minIONdict[minION]["APIHelp"].update_minion_current_stats(self.minIONdict[minION]['livedata'],
                                                                                          self.minIONdict[minION]['detailsdata'],
                                                                                          self.minIONdict[minION]["simplesummary"],
                                                                                          self.minIONdict[minION]["channelstuff"])
                                    except Exception as exception:
                                        print ("update failed")
                                        print(">>> Exception")
                                        print(exception)
                                        print("<<< Exception")


                                    self.minIONdict[minION]["APIHelp"].check_jobs(minION)

                            else:  # minION state is inactive
                                # print minION, "INACTIVE"
                                if minION in self.minIONclassdict:
                                    if "connected" in self.minIONclassdict[minION]:
                                        keys = self.minIONclassdict[minION].keys()
                                        for key in keys:
                                            #print (key)
                                            self.minIONclassdict[minION].pop(key, None)
                time.sleep(5)
            time.sleep(1)

    def hang_up(self):
        print ("ctrl-c detected at top level")
        if self.args.GUI:
            self.args.minKNOWmessage = "Disconnecting MinIONs from minoTour control."
        #print ("Connected to Master MinION Controller!")
        #print ("Disconnecting MinIONs from minoTour control.")
        for minION in self.minIONclassdict:
            self.minIONdict[minION]["APIHelp"].update_minion_status(minION, 'UNKNOWN', 'unplugged')
        print ("bye bye")



"""
File Routines for handling fastq files and monitoring locations. Built on watchdog.
"""
import logging
import os,sys
import threading
import time
import gzip
import numpy as np
import hashlib #for checking if a file is different to the value stored in the database

from tqdm import tqdm
from minFQ.minotourapiclient import Runcollection
from minFQ.minotourapi import MinotourAPI
from Bio import SeqIO
from watchdog.events import FileSystemEventHandler


log = logging.getLogger(__name__)

###Function modified from https://raw.githubusercontent.com/lh3/readfq/master/readfq.py

def readfq(fp): # this is a generator function
    last = None # this is a buffer keeping the last unprocessed line
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        desc, name, seqs, last = last[1:], last[1:].partition(" ")[0], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield desc,name, ''.join(seqs), None # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq): # have read enough quality
                    last = None
                    yield desc, name, seq, ''.join(seqs); # yield a fastq record
                    break
            if last: # reach EOF before reading enough quality
                yield desc, name, seq, None # yield a fasta record instead
                break


def md5Checksum(filePath):

    return os.path.getsize(filePath)

def check_fastq_path(path):
    folders = splitall(path)
    try:
        if folders[-2] in ("pass","fail"):
            return ("{}_{}".format(folders[-2],folders[-1]))
        elif folders[-3] in ("pass","fail"):
            return ("{}_{}_{}".format(folders[-3],folders[-2],folders[-1]))
        else:
            return ("{}".format(folders[-1]))
    except:
        return ("{}".format(folders[-1]))


def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def check_is_pass(path, avg_quality):

    folders = os.path.split(path)

    if 'pass' in folders[0]:

        return True

    elif 'fail' in folders[0]:

        return False

    else:
        if avg_quality != None:

            if avg_quality >= 7:

                return True  # This assumes we have been unable to find either pass or fail and thus we assume the run is a pass run.

            else:

                return False

        else:

            return True


def parse_fastq_description(description):
    """

    :param description:
    :return:
    """
    descriptiondict = dict()

    descriptors = description.split(" ")

    del descriptors[0]

    for item in descriptors:
        bits = item.split("=")
        descriptiondict[bits[0]] = bits[1]

    return descriptiondict


def parse_fastq_record(desc, name, seq, qual, fastq, rundict, args, header, fastqfile):

    log.debug("Parsing reads from file {}".format(fastq))

    fastq_read = {}

    description_dict = parse_fastq_description(desc)

    fastq_read['read'] = description_dict.get('read', None)

    fastq_read['runid'] = description_dict.get('runid', None)

    fastq_read['channel'] = description_dict.get('ch', None)

    fastq_read['start_time'] = description_dict.get('start_time', None)

    #fastq_read['is_pass'] = check_is_pass(fastq)
    fastq_read['read_id'] = name

    fastq_read['sequence_length'] = len(str(seq))

    fastq_read['fastqfile'] = fastqfile["id"]


    # get or create fastfile if not in dictionary?

    #fastq_read['fastqfilename'] = fastqfileid

    if fastq_read['runid'] not in rundict:

        rundict[fastq_read['runid']] = Runcollection(args, header)

        rundict[fastq_read['runid']].add_run(description_dict, args)

    rundict[fastq_read['runid']].get_readnames_by_run(fastqfile['id'])

    if fastq_read['read_id'] not in rundict[fastq_read['runid']].readnames:


        quality = qual


        ## Turns out this is not the way to calculate quality...
        #fastq_read['quality_average'] = quality_average = np.around([np.mean(np.array(list((ord(val) - 33) for val in quality)))], decimals=2)[0]

        if quality != None:
            fastq_read['quality_average'] = round(-10 * np.log10(np.mean(np.array(list( (10**(-(ord(val)-33)/10)) for val in quality )))),2)

        fastq_read['is_pass'] = check_is_pass(fastq,fastq_read['quality_average'])
        #print (quality_average)

        # use 'No barcode' for non-barcoded reads
        barcode_name = description_dict.get('barcode', None)

        if barcode_name:

            fastq_read['barcode_name'] = barcode_name

        else:

            fastq_read['barcode_name'] = 'No barcode'

        # add control-treatment if passed as argument
        if args.treatment_control:

            if int(fastq_read['channel']) % args.treatment_control == 0:

                fastq_read['barcode_name'] = fastq_read['barcode_name'] + ' - control'

            else:

                fastq_read['barcode_name'] = fastq_read['barcode_name'] + ' - treatment'

        # check if sequence is sent or not
        if args.skip_sequence:

            fastq_read['sequence'] = ''
            fastq_read['quality'] = ''

        else:

            fastq_read['sequence'] = str(seq)
            fastq_read['quality'] = qual

        rundict[fastq_read['runid']].add_read(fastq_read)

    else:
        args.reads_skipped += 1

def get_runid(fastq):
    if fastq.endswith(".gz"):
        with gzip.open(fastq, "rt") as file:
            for _ in range(1):
                line = file.readline()
    else:
        with open(fastq, "r") as file:
            for _ in range(1):
                line = file.readline()
    for _ in line.split():
        if _.startswith("runid"):
            # print (_.split("=")[1])
            runid = _.split("=")[1]
    return runid


def parse_fastq_file(fastq, rundict, fastqdict, args, header, MinotourConnection):

    log.debug("Parsing fastq file {}".format(fastq))

    runid=get_runid(fastq)


    fastqfile = MinotourConnection.create_file_info(str(check_fastq_path(fastq)), runid, "0", None)

    counter = 0

    if fastq.endswith(".gz"):

        with gzip.open(fastq, "rt") as fp:
            try:
                for desc, name, seq, qual in readfq(fp):
                    counter += 1

                    args.reads_seen += 1

                    args.fastqmessage = "processing read {}".format(counter)

                    parse_fastq_record(desc, name, seq, qual, fastq, rundict, args, header,fastqfile)

            except Exception as e:

                args.reads_corrupt += 1

                log.error("Corrupt file observed in {}.".format(fastq))
                log.error(e)
                #continue

    else:

        #for record in SeqIO.parse(fastq, "fastq"):
        with open(fastq, "r") as fp:

            for desc, name, seq, qual in readfq(fp):

                counter += 1

                args.reads_seen += 1

                args.fastqmessage = "processing read {}".format(counter)

                parse_fastq_record(desc, name, seq, qual, fastq, rundict, args, header, fastqfile)


        args.reads_corrupt += 1

        log.error("Corrupt file observed in {}.".format(fastq))

        #continue

    for runs in rundict:

        rundict[runs].commit_reads()

    try:
        fastqfile = MinotourConnection.create_file_info(str(check_fastq_path(fastq)), runid, md5Checksum(fastq), rundict[runid].run)
    except Exception as err:
        log.error("Problem with uploading file {}".format(err))

    return counter


def file_dict_of_folder_simple(path, args, MinotourConnection, fastqdict):

    file_list_dict = dict()

    if not args.ignoreexisting:
    
        counter = 0
    
        if os.path.isdir(path):

            log.info("caching existing fastq files in: %s" % (path))

            args.fastqmessage = "caching existing fastq files in: %s" % (path)

            novelrunset=set()

            seenfiletracker = dict()

            for path, dirs, files in os.walk(path):

                for f in files:

                    if f.endswith(".fastq") or f.endswith(".fastq.gz"):
                        log.debug("Processing File {}\r".format(f))
                        counter += 1

                        args.files_seen += 1
                        #### Here is where we want to check if the files have been created and what the checksums are
                        #### If the file checksums do not match, we pass the file to the rest of the script.
                        #### When we finish analysing a file, we will need to update this information n the server.
                        #### Currently just using size.
                        md5Check = md5Checksum(os.path.join(path, f))

                        runid = get_runid(os.path.join(path, f))
                        if runid not in novelrunset and runid not in seenfiletracker.keys():
                            result = (MinotourConnection.get_file_info_by_runid(runid))
                            #### Here we want to parse through the results and store them in some kind of dictionary in order that we can check what is happening
                            if result is not None:
                                for entry in result:
                                    if entry["runid"] not in seenfiletracker.keys():
                                        seenfiletracker[entry["runid"]]=dict()
                                    seenfiletracker[entry["runid"]][entry["name"]]=entry["md5"]
                        else:
                            result = None

                        filepath = os.path.join(path, f)
                        checkfilepath = check_fastq_path(filepath)
                        if checkfilepath not in fastqdict.keys():
                            fastqdict[checkfilepath]=dict()

                        fastqdict[checkfilepath]["runid"] = runid
                        fastqdict[checkfilepath]["md5"] = md5Check


                        """Here we are going to check if the files match or not. """
                        seenfile = False
                        if runid in seenfiletracker.keys() and checkfilepath in seenfiletracker[runid].keys():
                            seenfile = True
                            if int(md5Check) == int(seenfiletracker[runid][checkfilepath]):
                                args.files_skipped += 1
                            else:
                                file_list_dict[filepath] = os.stat(filepath).st_mtime
                                novelrunset.add(runid)

                        if not seenfile:
                            file_list_dict[filepath] = os.stat(filepath).st_mtime
                            novelrunset.add(runid)

        log.info("processed %s files" % (counter))

        args.fastqmessage = "processed %s files" % (counter)

        log.info("found %d existing fastq files to process first." % (len(file_list_dict)))

    else:
        args.fastqmessage = "Ignoring existing fastq files in: %s" % (path)
    
    return file_list_dict


class FastqHandler(FileSystemEventHandler):

    def __init__(self, args, header, rundict):
        """
        Collect information about files already in the folders
        """

        self.file_descriptor = dict()
        self.args = args
        self.header = header
        self.args.files_seen = 0
        self.args.files_processed = 0
        self.args.files_skipped = 0
        self.args.reads_seen = 0
        self.args.reads_corrupt = 0
        self.args.reads_skipped = 0
        self.args.reads_uploaded = 0
        # adding files to the file_descriptor is really slow - therefore lets skip that and only update the files when we want to basecall thread_number
        self.MinotourConnection = MinotourAPI(self.args.host_name, self.args.port_number, self.header)
        self.rundict = rundict
        self.fastqdict= dict()
        #if not self.args.ignoreexisting:
        self.creates = file_dict_of_folder_simple(self.args.watchdir, self.args, self.MinotourConnection,self.fastqdict)
        self.processing = dict()
        self.running = True


        self.t = threading.Thread(target=self.processfiles)
        self.grouprun = None

        try:
            self.t.start()
        except KeyboardInterrupt:
            self.t.stop()
            raise

    def addrunmonitor(self,runpath):
        """
        Add a new folder for checking reads in.
        :param runpath: the final part of the folder structure - effectively the sample name
        :return:
        """
        pass

    def stopt(self):
        self.running=False

    def lencreates(self):
        return len(self.creates)

    def lenprocessed(self):
        return len(self.processed)

    def processfiles(self):
        #print ("Process Files Inititated")
        while self.running:
            currenttime = time.time()
            #for fastqfile, createtime in tqdm(sorted(self.creates.items(), key=lambda x: x[1])):
            for fastqfile, createtime in sorted(self.creates.items(), key=lambda x: x[1]):

                delaytime = 10

                # file created 5 sec ago, so should be complete. For simulations we make the time longer.
                if (int(createtime) + delaytime < time.time()):

                    del self.creates[fastqfile]

                    #print (fastqfile,md5Checksum(fastqfile), "\n\n\n\n")

                    parse_fastq_file(fastqfile, self.rundict, self.fastqdict, self.args, self.header, self.MinotourConnection)

                    self.args.files_processed += 1

            if currenttime+5 > time.time():
                time.sleep(5)
            #print ("still ticking")

    def process_fastqfile(self, filename):

        parse_fastq_file(filename, self.rundict, self.fastqdict, self.args, self.header, self.MinotourConnection)

    def on_created(self, event):
        """Watchdog counts a new file in a folder it is watching as a new file"""
        """This will add a file which is added to the watchfolder to the creates and the info file."""
        # if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz")):
        #     self.creates[event.src_path] = time.time()

        log.info("Processing file {}".format(event.src_path))
        #time.sleep(5)
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(".fq") or event.src_path.endswith(".fq.gz")):
            self.args.files_seen += 1
            self.creates[event.src_path] = time.time()

    def on_modified(self, event):
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(".fq") or event.src_path.endswith(".fq.gz")):
            log.debug("Modified file {}".format(event.src_path))
            self.creates[event.src_path] = time.time()
        # elif (event.dest_path.endswith(".fastq") or event.dest_path.endswith(".fastq,gz") or event.dest_path.endswith(".gq") or event.dest_path.endswith(".fq.gz")):
        #     log.debug("Modified file {}".format(event.dest_path))
        #     self.creates[event.dest_path] = time.time()
        
    def on_moved(self, event):
        if any((event.dest_path.endswith(".fastq"), event.dest_path.endswith(".fastq,gz"), event.dest_path.endswith(".gq"), event.dest_path.endswith(".fq.gz"))):
            log.debug("Modified file {}".format(event.dest_path))
            self.creates[event.dest_path] = time.time()

#self.process_fastqfile(event.src_path)

        # f = open(event.src_path, "r")
        # counter = 0
        # for line in f:
        #     log.info("{} - {}".format(event.src_path, counter))
        #     counter = counter + 1



"""
A class to handle the collection of run statistics and information 
from fastq files and upload to minotour.
"""
import datetime
import json
import logging
import sys

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from urllib.parse import urlparse
from threading import Thread


log = logging.getLogger(__name__)

# https://www.peterbe.com/plog/best-practice-with-retries-with-requests
def requests_retry_session(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
    session=None,
):

    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session

from minFQ.minotourapi import MinotourAPI


class Runcollection():

    def __init__(self, args, header):

        log.debug("Initialising Runcollection")

        self.base_url = args.host_name
        self.port_number = args.port_number
        self.check_url()
        self.args = args
        self.header = header
        self.readnames = list()
        self.readcount = 0
        self.basecount = 0
        self.read_type_list = dict()
        if self.args.skip_sequence:
            self.batchsize = 5000
        else:
            self.batchsize = 100
        self.run = None
        self.grouprun = None
        self.barcode_dict = {}
        self.read_list = []
        self.filemonitor = dict()
        self.fastqfileid = None
        self.minotourapi = MinotourAPI(self.args.host_name, self.args.port_number, self.header)
        self.get_readtype_list()


    def check_url(self):
        if self.base_url.startswith("http://"):
            self.base_url = self.base_url[7:]
        if self.base_url.startswith(("https://")):
            self.base_url = self.base_url[8:]
        if int(self.port_number) != 80:
            r = requests.get("http://{}:{}/".format(self.base_url, self.port_number))
        else:
            r = requests.get("http://{}/".format(self.base_url))
        #print (r.url)
        if r.url.startswith("https"):
            self.base_url="https://{}/".format(self.base_url)
        else:
            self.base_url="http://{}:{}/".format(self.base_url,self.port_number)

    def get_readtype_list(self):

        read_type_list = self.minotourapi.get_read_type_list()

        read_type_dict = {}

        for read_type in read_type_list:

            read_type_dict[read_type["name"]] = read_type["id"]

        self.read_type_list = read_type_dict


    def get_readnames_by_run(self,fastqfileid):

        if fastqfileid != self.fastqfileid:
            self.fastqfileid = fastqfileid
            # TODO move this function to MinotourAPI class.

            #url = "{}api/v1/runs/{}/readnames/".format(self.base_url, self.run['id'])
            url = "{}api/v1/runs/{}/readnames/".format(self.base_url, fastqfileid)

            req = requests.get(
                url,
                headers=self.header
            )

            readname_list = json.loads(req.text)

            number_pages = readname_list['number_pages']

            log.debug("Fetching reads to check if we've uploaded these before.")
            log.debug("Wiping previous reads seen.")
            self.readnames = list()
            #for page in tqdm(range(number_pages)):
            for page in range(number_pages):

                self.args.fastqmessage = "Fetching {} of {} pages.".format(page,number_pages)

                new_url = url + '?page={}'.format(page)

                content = requests.get(new_url,  headers=self.header)

                log.debug("Requesting {}".format(new_url))

                # We have to recover the data component and loop through that to get the read names.
                for read in json.loads(content.text)["data"]:

                    self.readnames.append(read)

            log.debug("{} reads already processed and included into readnames list for run {}".format(len(self.readnames), self.run['id']))

    def add_run(self, descriptiondict, args):

        self.args.fastqmessage = "Adding run."

        runid = descriptiondict["runid"]

        run = self.minotourapi.get_run_by_runid(runid)

        if not run:

            #print (descriptiondict)

            if 'flow_cell_id' in descriptiondict:
                flowcellname = descriptiondict['flow_cell_id']
            elif 'sampleid' in descriptiondict:
                flowcellname = descriptiondict['sampleid']
            else:
                flowcellname = self.args.run_name


            #
            # get or create a flowcell
            #

            log.debug("Looking for flowcell {}".format(flowcellname))

            flowcell = self.minotourapi.get_flowcell_by_name(flowcellname)
            log.info("FLOWCELL IS")
            log.info(flowcell)
            if flowcell is not None:
                flowcell = flowcell["data"]

            log.debug("found {}".format(flowcell))

            if flowcell is None:

                log.debug("Trying to create flowcell {}".format(flowcellname))

                flowcell = self.minotourapi.\
                    create_flowcell(flowcellname)
                # print(dir(args))
                # If we have a job as an option
                if args.job:
                    # If there is a target set
                    if args.targets is not None:
                        self.minotourapi.create_job(flowcell['id'], int(args.job_id), None, args.targets)
                    # If there is a reference and not a target set
                    elif args.reference and not args.targets:
                        self.minotourapi.create_job(flowcell['id'], int(args.job_id), args.reference, None)
                    # If there is neither
                    else:
                        self.minotourapi.create_job(flowcell['id'], int(args.job_id), None, None)
                log.debug("Created flowcell {}".format(flowcellname))

            #
            # create a run
            #
            if "barcode" in descriptiondict.keys():

                is_barcoded = True

            else:

                is_barcoded = False

            if self.args.skip_sequence:

                has_fastq = False

            else:

                has_fastq = True

            if 'sample_id' in descriptiondict:
                runname = descriptiondict['sample_id']
            else:
                runname = self.args.run_name

            createrun = self.minotourapi.create_run(runname, runid, is_barcoded, has_fastq, flowcell)

            if not createrun:

                log.critical('There is a problem creating run')
                sys.exit()

            else:
                log.debug("run created")

            run = createrun

        if not self.run:

            self.run = run

        for item in run['barcodes']:

            self.barcode_dict.update({
                item['name']: item['id']
            })

        #self.get_readnames_by_run()

    def commit_reads(self):

        #Thread(target=self.minotourapi.create_reads(self.read_list)).start()
        self.minotourapi.create_reads(self.read_list)
        self.args.reads_uploaded += len(self.read_list)

        self.read_list = list()

    def update_read_type(self, read_id, type):
        payload = {'type': type}
        updateread = requests.patch(self.args.full_host + 'api/v1/runs/' + str(self.runid) + "/reads/" + str(read_id) + '/',
                                    headers=self.header, json=payload)

    def check_1d2(self, readid):
        if len(readid) > 64:
            return True

    def add_read(self, fastq_read_payload):

        barcode_name = fastq_read_payload['barcode_name']
        runid = fastq_read_payload['runid']
        read_id = fastq_read_payload['read_id']

        fastq_read_payload['run'] = self.run['id']

        if barcode_name not in self.barcode_dict:

            # print(">>> Found new barcode {} for run {}.".format(barcode_name, runid))

            barcode = self.minotourapi.create_barcode(barcode_name, self.run['url'])

            if barcode:

                self.barcode_dict.update({
                    barcode['name']: barcode['id']
                })

            else:
                log.critical("Problem finding barcodes.")
                sys.exit()

        fastq_read_payload['barcode'] = self.barcode_dict[fastq_read_payload['barcode_name']]

        if read_id not in self.readnames:

            if self.check_1d2(read_id):

                firstread, secondread = read_id[:len(read_id) // 2], read_id[len(read_id) // 2:]

                self.update_read_type(secondread, self.read_type_list["C"])

                fastq_read_payload['type'] = self.read_type_list["1D^2"]

            else:

                fastq_read_payload['type'] = self.read_type_list["T"]

            self.readnames.append(read_id)

            if self.args.GUI:
                self.args.readcount += 1

            if self.args.GUI:
                self.args.basecount += fastq_read_payload['sequence_length']
            self.basecount += fastq_read_payload['sequence_length']

            if self.args.GUI:
                self.args.qualitysum += fastq_read_payload['quality_average']

            self.read_list.append(fastq_read_payload)

            log.debug('Checking read_list size {} - {}'.format(len(self.read_list), self.batchsize))

            if len(self.read_list) >= self.batchsize:
                if not self.args.skip_sequence:
                    self.batchsize = int(1000000/(int(self.basecount)/self.readcount))
                self.commit_reads()

            self.readcount += 1

        else:
            print ("Skipping read")
            self.args.reads_skipped += 1


import datetime
import json as jsonlibrary
import logging
import requests
import gzip
import time

from urllib.parse import urlparse


log = logging.getLogger(__name__)




class MinotourAPI:

    def __init__(self, base_url, port_number, request_headers):

        self.base_url = base_url
        self.port_number = port_number
        self.check_url()
        self.request_headers = request_headers
        self.minion_event_types = self.get_minion_event_types()

    def check_url(self):
        if self.base_url.startswith("http://"):
            self.base_url = self.base_url[7:]
        if self.base_url.startswith(("https://")):
            self.base_url = self.base_url[8:]
        if int(self.port_number) != 80:
            r = requests.get("http://{}:{}/".format(self.base_url, self.port_number))
        else:
            r = requests.get("http://{}/".format(self.base_url))
        #print (r.url)
        if r.url.startswith("https"):
            self.base_url="https://{}/".format(self.base_url)
        else:
            self.base_url="http://{}:{}/".format(self.base_url,self.port_number)


    def test(self):

        return "OK"

    def get(self, partial_url, parameters=None):

        if not parameters:

            url = '{}api/v1{}'.format(self.base_url, partial_url)

        else:

            url = '{}api/v1{}'.format(self.base_url, partial_url)

        return requests.get(url, headers=self.request_headers, params=parameters)

    def post(self, partial_url, json, parameters=None):

        if not parameters:

            url = '{}api/v1{}'.format(self.base_url, partial_url)

        else:

            url = '{}api/v1{}?{}'.format(self.base_url, partial_url, parameters)

        #print (url)

        return requests.post(url, headers=self.request_headers, json=json)

    def put(self, partial_url, json, parameters=None):

        if not parameters:

            url = '{}api/v1{}'.format(self.base_url, partial_url)

        else:

            url = '{}api/v1{}?{}'.format(self.base_url, partial_url, parameters)

        return requests.put(url, headers=self.request_headers, json=json)

    def delete(self, partial_url, json, parameters=None):

        if not parameters:

            url = '{}api/v1{}'.format(self.base_url, partial_url)

        else:

            url = '{}api/v1{}?{}'.format(self.base_url, partial_url, parameters)

        return requests.delete(url, headers=self.request_headers, json=json)

    def get_target_sets(self, api_key=""):
        """
        Get a list of the target sets, to show during the listing
        :return:
        """
        url = '/metagenomics/targetsets'
        if api_key == "":
            payload = {}
        else:
            payload = {"api_key": api_key, "cli": True}

        req = self.get(url, parameters=payload)

        if req.status_code != 200:
            log.error("Couldn't find target sets to run.")
            log.error("Status-code {}".format(req.status_code))
            log.error("Text {}".format(req.text))
            return None

        else:
            return jsonlibrary.loads(req.text)


    def get_job_options(self):
        """
        Get a list of the jobs available to be started from the client
        :return:
        """
        url = '/tasktypes/'

        payload = {"cli": True}

        req = self.get(url, parameters=payload)

        if req.status_code != 200:
            log.error("Couldn't find tasks to run.")
            log.error("Status-code {}".format(req.status_code))
            log.error("Text {}".format(req.text))
            return None

        else:
            return jsonlibrary.loads(req.text)

    def get_references(self):

        url = '/reference/'

        req = self.get(url)

        if req.status_code != 200:
            log.error("Couldn't find references.")
            log.error("Status-code {}".format(req.status_code))
            log.error("Text {}".format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)


    def get_file_info_by_runid(self, runid):

        url = '/runs/{}/files/'.format(runid)

        req =self.get(url)

        if req.status_code != 200:
            log.error('Did not find files for {}.'.format(runid))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def create_file_info(self, filename, runid, md5check, run):

        payload = {
            'name': filename,
            'runid': runid,
            'md5' : md5check,
            'run' : run
        }

        url = '/runs/{}/files/'.format(runid)

        req = self.post(url, json=payload)

        if req.status_code != 201:

            log.error('File info {} could not be created.')
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:
            fileinfo = jsonlibrary.loads(req.text)
            return fileinfo

    def get_run_by_runid(self, runid):

        url = '/runs/{}/'.format(runid)

        req = self.get(url, 'search_criteria=runid')

        if req.status_code != 200:

            log.info('Did not find run {}.'.format(runid))
            log.info('Creating run {}.'.format(runid))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def get_grouprun_by_name(self, name):

        url = '/grouprun/{}/'.format(name)

        req = self.get(url, 'search_criteria=name')

        if req.status_code != 200:

            log.error('Did not find grouprun {}.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            grouprun = jsonlibrary.loads(req.text)
            return grouprun

    def create_run(self, name, runid, is_barcoded, has_fastq, flowcell, minion = None, start_time = None):
        
        payload = {
            "name": name,
            "sample_name": name,
            "runid": runid,
            "is_barcoded": is_barcoded,
            "has_fastq": has_fastq,
            "flowcell": flowcell['url'],
        }

        if minion:
            payload["minion"]=minion['url']

        if start_time:
            payload["start_time"]=start_time

        req = self.post('/runs/', json=payload)

        if req.status_code != 201:

            log.error('Run {} could not be created.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            run = jsonlibrary.loads(req.text)
            return run

    def create_grouprun(self, name):

        payload = {
            'name': name
        }

        req = self.post('/grouprun/', json=payload)

        if req.status_code != 201:

            log.error('GroupRun {} could not be created.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            grouprun = jsonlibrary.loads(req.text)
            return grouprun

    def create_grouprun_membership(self, grouprun, run):

        payload = {
            'grouprun_id': grouprun,
            'run_id': run
        }

        req = self.post('/grouprun-membership/', json=payload)

        if req.status_code != 201:

            log.error('GroupRun membership {} could not be created.')
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            grouprun = jsonlibrary.loads(req.text)
            return grouprun

    def create_reads(self, reads):

        #log.debug('Creating reads')
        #time.sleep(1)
        payload = reads

        req = self.post('/read/', json=payload)

        if req.status_code != 201:

            log.error('Reads batch {} could not be created.')
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            grouprun = jsonlibrary.loads(req.text)
            return grouprun

    def get_read_type_list(self):

        url = '/readtypes/'

        req = self.get(url)

        if req.status_code != 200:

            log.error('FastqReadType list could not be retrieved.')
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            read_type_list = jsonlibrary.loads(req.text)
            return read_type_list

    def create_barcode(self, barcode_name, run_url):

        payload = {
            'name': barcode_name,
            'run': run_url
        }

        req = self.post('/barcode/', json=payload)

        if req.status_code != 201:

            log.error('Barcode {} could not be created.'.format(barcode_name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            barcode = jsonlibrary.loads(req.text)
            return barcode

    def get_flowcell_by_name(self, name):

        url = '/flowcells/{}/'.format(name)

        req = self.get(url, 'search_criteria=name')

        if req.status_code == 200:

            return jsonlibrary.loads(req.text)

        elif req.status_code == 404:

            return {'data':{}}

        else:

            log.error('Did not find flowcell {}.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Unrecognised response fetching Flowcell from minoTour')
            log.error('Text {}'.format(req.text))
            return None


    def create_flowcell(self, name):

        payload = {

            'name': name,
        }

        req = self.post('/flowcells/', json=payload)

        if req.status_code != 201:

            log.error('Flowcell {} could not be created.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def create_job(self, flowcell, job, reference=None, targets=None):
        """

        :param flowcell:
        :param job:
        :param reference:
        :param targets:
        :return:
        """
        payload = {
            'flowcell': flowcell,
            'job_type': job
        }

        if reference is not None:
            payload['reference'] = reference

        if targets is not None:
            payload["target_set"] = targets

        req = self.post('/tasks/', json=payload)

        if req.status_code != 200:
            log.info(req.status_code)
            log.info(req.text)

    def get_minion_by_name(self, name):

        url = '/minions/{}/'.format(name)

        req = self.get(url, 'search_criteria=name')

        if req.status_code != 200:

            log.debug('Did not find minion {}.'.format(name))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def create_minion(self, name):

        payload = {

            'minION_name': name,
            'name': name,
        }

        req = self.post('/minions/', json=payload)

        if req.status_code != 201:

            log.error('Minion {} could not be created.'.format(name))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def get_server_version(self):

        url = '/version'

        req = self.get(url)

        if req.status_code != 200:

            log.error('Did not find version for server {}.'.format(self.base_url))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    ### New functions being added for minknowconnection use.
    def identify_minion(self,minION):
        #print ("args full host {}".format(self.args.full_host))
        r = requests.get(self.args.full_host + 'api/v1/minions', headers=self.header)
        minionidlink = ""
        for minion in jsonlibrary.loads(r.text):
            if minion["minION_name"] == minION:
                minionidlink = minion["url"]
        return(minionidlink)

    def update_minion(self,minION):
        """
        Receives a MinION object from minknowconnection and passes it to the api to
        update any values within it.
        :param minION:
        :return:
        """

        url = '/minions/{}/'.format(minion.id)

        req = self.post(url, json=minion)

        if req.status_code != 201:

            log.debug('Minion {} could not be updated.'.format(name))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def get_minion_event_types(self):

        url = '/minioneventtypes/'

        req = self.get(url)

        if req.status_code != 200:

            log.error('Did not find version for server {}.'.format(self.base_url))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)


    def fetch_minion_scripts(self,minion):

        url = '/minions/{}/scripts/'.format(minion['id'])

        req = self.get(url)

        if req.status_code != 200:
            log.error('Did not find scripts for minion {}.'.format(minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            self.minIONscripts = None  # TODO we should move this code to minknowconnection
            return None

        else:
            minIONscripts = jsonlibrary.loads(req.text)
            scriptidlist = list()
            for script in minIONscripts:
                scriptidlist.append(script["identifier"])
            self.minIONscripts = scriptidlist  # TODO we should move this code to minknowconnection
            #print (self.minIONscripts)
            return jsonlibrary.loads(req.text)


    def update_minion_script(self,minion,scriptdictionary):
        """

        :param minion:
        :param scriptdictionary:
        :return:
        """

        url = '/minions/{}/scripts/'.format(minion['id'])

        if scriptdictionary["identifier"] not in self.minIONscripts:  # TODO we should move this logic to minknowconnection
            #print ("!!!!!Updating Scripts!!!!!")
            payload = dict()

            for item in scriptdictionary:
                payload[item.replace(" ", "_")]=scriptdictionary[item]

            payload["minION"]=str(minion["url"])

            req = self.post(url, json=payload)

            if req.status_code != 201:

                log.debug('script {} could not be updated.'.format(payload))
                log.debug('Status-code {}'.format(req.status_code))
                log.debug('Text {}'.format(req.text))
                return None

            else:

                return jsonlibrary.loads(req.text)

            pass

        #else:
            #print ("not updating")
            #print (scriptdictionary)


    def update_minion_event(self,minion,computer,status):
        """

        :param minION: minION
        :param computer: computer_name
        :param status: event
        :return:
        """

        url = '/minions/{}/events/'.format(minion['id'])
        for info in self.minion_event_types:
            for item in info:

                if item == "name":
                    if info[item]==status:
                        statusidlink= info["url"]

        payload = {
            "computer_name": computer, 
            "datetime": str(datetime.datetime.now()), 
            "event": str(urlparse(statusidlink).path),
            "minION": str(minion["url"])
        }

        req = self.post(url, json=payload)

        if req.status_code != 201:

            log.debug('event {} could not be updated.'.format(status))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def update_minion_run_info(self,payload,runid):

        payload = payload

        url = '/runs/{}/rundetails/'.format(runid)

        req = self.post(url, json=payload)

        if req.status_code != 201:
            #print (req.text)
            log.debug('runid {} could not be updated.'.format(runid))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def get_minion_status(self,minion):

        url = '/minions/{}/status/'.format(minion['id'])

        req = self.get(url)

        if req.status_code != 200:

            log.error('minion ID {} could not be checked.'.format(minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass


    def create_minion_status(self, payload, minion):

        payload = payload

        url = '/minions/{}/status/'.format(minion['id'])

        req = self.post(url, json=payload)

        if req.status_code != 201:

            log.error('minion status {} could not be created.'.format(minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def update_minion_status(self,payload,minion):

        payload = payload

        url = '/minions/{}/status/'.format(minion['id'])

        req = self.put(url, json=payload)

        if req.status_code != 200:

            log.error('minion status {} could not be updated.'.format(minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def create_minion_statistic(self,payload,runid):

        url = '/runs/{}/runstats/'.format(runid)

        req = self.post(url, json=payload)

        if req.status_code != 201:

            log.error('minion statistic {} could not be updated.'.format(runid))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def create_message(self, payload, minion):

        payload = payload

        url = '/minions/{}/messages/'.format(minion['id'])

        req = self.post(url, json=payload)

        if req.status_code not in  (200,201):

            log.debug('minion message {} could not be updated.'.format(minion['id']))
            log.debug('Status-code {}'.format(req.status_code))
            log.debug('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

        pass

    def get_minion_jobs(self, minion):

        url = '/minions/{}/control/'.format(minion['id'])

        req = self.get(url)

        if req.status_code != 200:

            log.error('Did not find jobs for minion {}.'.format(self.minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            return jsonlibrary.loads(req.text)

    def complete_minion_job(self,minion,job):

        url = '/minions/{}/control/{}/'.format(minion['id'],job['id'])

        req = self.post(url,json=None)

        if req.status_code != 204:

            log.error('Could not complete job for minion {}.'.format(minion['id']))
            log.error('Status-code {}'.format(req.status_code))
            log.error('Text {}'.format(req.text))
            return None

        else:

            pass



import datetime
import json
import logging
import sys
import threading
import time

import configargparse
import pandas as pd
from google.protobuf.json_format import MessageToJson
from ws4py.client.threadedclient import WebSocketClient

import minFQ.rpc as rpc
from minFQ.minotourapi import MinotourAPI as MinotourAPINew
from minFQ.minotourutils import determinetype, parse_ports

log = logging.getLogger(__name__)

def parsemessage(message):
    return json.loads(MessageToJson(message, preserving_proto_field_name=True, including_default_value_fields=True))

import grpc

rpc._load()
import minFQ.rpc.protocol_pb2 as protocol
import minFQ.rpc.protocol_pb2_grpc as protocol_grpc
import minFQ.rpc.manager_pb2 as manager
import minFQ.rpc.manager_pb2_grpc as manager_grpc

#from minotourAPI import

class DeviceConnect(WebSocketClient):
    #def __init__(self, *args,**kwargs):
    #    super(DeviceConnect, self).__init__(*args,**kwargs)
        #self.detailsdict=dict()
    #    self.daemon=True
    def __init__(self, connectip,args,rpcconnection,header,minIONid):
        self.args = args
        if self.args.verbose:
            log.info("Client established!")
        WebSocketClient.__init__(self, connectip)
        self.rpc_connection=rpcconnection

        #Here we need to check if we are good to run against this version.
        self.version = self.rpc_connection.instance.get_version_info().minknow.full
        self.device_type = parsemessage(self.rpc_connection.instance.get_host_type())['host_type']
        #log.error(self.device_type)
        if str(self.device_type).startswith("PROMETHION"):
            log.warning(self.device_type)
            log.warning("This version of minFQ may not be compatible with PromethION.")
            #sys.exit()
        #if str(self.version) != "3.3.13":
        if not str(self.version).startswith("3.3"):
            log.warning(self.version)
            log.warning("This version of minFQ may not be compatible with the MinKNOW version you are running.")
            log.warning("As a consequence, live monitoring MAY NOT WORK.")
            log.warning("If you experience problems, let us know.")
            #sys.exit()
        self.header=header
        self.channels = parsemessage(self.rpc_connection.device.get_flow_cell_info())['channel_count']
        self.channelstatesdesc = self.rpc_connection.analysis_configuration.get_channel_states_desc()
        self.channelstates = dict()
        for i in range(self.channels):
            self.channelstates[i+1]=None
        self.status = ""
        self.interval = 30 #we will poll for updates every 30 seconds.
        self.longinterval = 30 #we have a short loop and a long loop
        self.minIONid = minIONid
        self.computer_name = self.rpc_connection.instance.get_machine_id().machine_id
        self.minknow_version = self.rpc_connection.instance.get_version_info().minknow.full
        self.minknow_status = self.rpc_connection.instance.get_version_info().protocols
        self.minotourapi = MinotourAPINew(self.args.host_name,self.args.port_number, self.header)
        self.minotourapi.test()
        self.disk_space_info = json.loads(
            MessageToJson(self.rpc_connection.instance.get_disk_space_info(), preserving_proto_field_name=True,
                          including_default_value_fields=True))
        self.flowcelldata = parsemessage(self.rpc_connection.device.get_flow_cell_info())
        minion = self.minotourapi.get_minion_by_name(self.minIONid)
        if not minion:
            minion = self.minotourapi.create_minion(self.minIONid)
        self.minion = minion
        self.minIONstatus = self.minotourapi.get_minion_status(self.minion)
        self.runidlink=""

        try:
            self.acquisition_data = parsemessage(self.rpc_connection.acquisition.get_acquisition_info())
        except:
            #if self.args.verbose:
            log.debug("No active run")
            self.acquisition_data = {}

        runmonitorthread = threading.Thread(target=self.runmonitor, args=())
        runmonitorthread.daemon = True                            # Daemonize thread
        runmonitorthread.start()

        flowcellmonitorthread = threading.Thread(target=self.flowcellmonitor, args=())
        flowcellmonitorthread.daemon = True
        flowcellmonitorthread.start()

        runinforthread = threading.Thread(target=self.runinfo, args=())
        runinforthread.daemon = True  # Daemonize thread
        runinforthread.start()



        messagesmonitor = threading.Thread(target=self.getmessages, args=())
        messagesmonitor.daemon = True  # Daemonize thread
        messagesmonitor.start()

        #This is for future usage.
        #dutytimemonitorthread = threading.Thread(target=self.dutytimemonitor, args=())
        #dutytimemonitorthread.daemon = True
        #dutytimemonitorthread.start()

        newchannelstatethread = threading.Thread(target=self.newchannelstatemonitor, args=())
        newchannelstatethread.daemon = True
        newchannelstatethread.start()

        newhistogrammonitorthread = threading.Thread(target=self.newhistogrammonitor, args=())
        newhistogrammonitorthread.daemon = True
        newhistogrammonitorthread.start()

        jobsmonitorthread = threading.Thread(target=self.jobs_monitor,args=())
        jobsmonitorthread.daemon = True
        jobsmonitorthread.start()


        log.debug("All is well with connection.")
        self.first_connect()

    def disconnect_nicely(self):
        log.debug("Trying to disconnect nicely")
        self.minotourapi.update_minion_event(self.minion, self.computer_name, "unplugged")
        try:
            self.minIONstatus["minKNOW_status"]="unplugged"
        except:
            log.debug("Couldn't unplug MinION from website.")
        self.minIONstatus = self.minotourapi.update_minion_status(self.minIONstatus, self.minion)

    def first_connect(self):
        """
        This function will run when we first connect to the MinION device.
        It will provide the information to minotour necessary to remotely control the minION device.
        :return:
        """
        log.debug("First connection observed")
        log.debug("All is well with connection. {}".format(self.minion))
        self.minotourapi.update_minion_event(self.minion,self.computer_name,"active")
        self.minotourapi.fetch_minion_scripts(self.minion)
        for protocol in self.rpc_connection.protocol.list_protocols().ListFields()[0][1]:
            protocoldict = self.parse_protocol(protocol)
            #print (self.minion,protocoldict)
            self.minotourapi.update_minion_script(self.minion,protocoldict)
        if str(self.status).startswith("status: PROCESSING"):
            self.run_start()

    def parse_protocol(self,protocol):
        protocoldict=dict()
        flowcell = "N/A"
        kit = "N/A"
        basecalling = "N/A"
        try:
            kit = protocol.tags["kit"].ListFields()[0][1]
            flowcell = protocol.tags["flow cell"].ListFields()[0][1]
            basecalling = protocol.tags["base calling"].ListFields()[0][1]
        except:
            pass
        #print (protocol.name, protocol.identifier)
        protocoldict["identifier"]=protocol.identifier
        #protocoldict["name"]=protocol.name
        #print (protocol.name)
        if basecalling:
            basecalling="BaseCalling"
        else:
            basecalling="NoBaseCalling"
        protocoldict["name"]="{}/{}_{}_{}".format(protocol.name,flowcell,kit,basecalling)
        for tag in protocol.tags:
            try:
                protocoldict[tag]=protocol.tags[tag].ListFields()[0][1]
            except:
                pass
        return protocoldict

    def run_start(self):
        """
        This function will fire when a run first begins.
        It will drive the creation of a run.
        :return:
        """
        self.minotourapi.update_minion_event(self.minion, self.computer_name, "sequencing")

        log.debug("run start observed")
        log.debug("MINION:", self.minion)
        #We wait for 10 seconds to allow the run to start
        time.sleep(self.interval)
        try:
            self.runinformation = self.rpc_connection.acquisition.get_current_acquisition_run()

            log.debug(self.runinfo_api)
            log.debug(self.sampleid)
            log.debug(self.runinformation)
            log.debug("RUNID", self.runinformation.start_time)
            log.debug(self.channelstatesdesc)
            log.debug(self.channels)
            log.debug("FLOWCELL DATA", self.get_flowcell_id())
            log.debug("trying to create run")
            self.create_run(self.runinformation.run_id)
            log.debug("run created!!!!!!!")
            self.update_minion_run_info()
            log.debug("update minion run info complete")

        except Exception as err:
            log.error("Problem:", err)



    def update_minion_run_info(self):
        payload = {
            "minION": str(self.minion["url"]),
            "minKNOW_current_script": str(self.rpc_connection.protocol.get_run_info().protocol_id),
            "minKNOW_sample_name": str(self.sampleid.sample_id),
            "minKNOW_exp_script_purpose": str(self.rpc_connection.protocol.get_protocol_purpose()),
            "minKNOW_flow_cell_id": self.get_flowcell_id(),
            "minKNOW_run_name": str(self.sampleid.sample_id),
            "run_id": str(self.runidlink),
            "minKNOW_version": str(self.rpc_connection.instance.get_version_info().minknow.full),
            "minKNOW_hash_run_id": str(self.runinformation.run_id),
            "minKNOW_script_run_id": str(self.rpc_connection.protocol.get_current_protocol_run().acquisition_run_ids[0]),
            "minKNOW_real_sample_rate": int(str(self.rpc_connection.device.get_sample_rate().sample_rate            )),
            "minKNOW_asic_id": self.flowcelldata['asic_id'],
            "minKNOW_start_time": self.runinformation.start_time.ToDatetime().strftime('%Y-%m-%d %H:%M:%S'),
            #"minKNOW_colours_string": str(self.rpc_connection.analysis_configuration.get_channel_states_desc()),
            "minKNOW_colours_string": str(MessageToJson(self.rpc_connection.analysis_configuration.get_channel_states_desc(), preserving_proto_field_name=True, including_default_value_fields=True)),
            "minKNOW_computer": str(self.computer_name),
        }

        contextinfo = parsemessage(self.rpc_connection.protocol.get_context_info())['context_info']
        for k, v in contextinfo.items():
            payload[k]=v

        ruinfo = parsemessage(self.rpc_connection.protocol.get_run_info())

        try:
            payload['experiment_id']=ruinfo['user_info']['protocol_group_id']
        except:
            payload['experiment_id']="Not Known"

        log.debug(">>>>>>>>", payload)
        updateruninfo = self.minotourapi.update_minion_run_info(payload,self.runid)
        log.debug(updateruninfo)

    def create_run(self, runid):
        log.debug(">>> inside create_run")

        log.debug(self.minotourapi)

        log.debug(">>> after self.minotourapi")

        self.minotourapi.test()

        log.debug(">>> after self.minotourapi.test()")

        run = self.minotourapi.get_run_by_runid(runid)

        log.debug(run)

        if not run:
            log.debug(">>> no run {}".format(runid))
            #
            # get or create a flowcell
            #
            flowcell = self.minotourapi.get_flowcell_by_name(self.get_flowcell_id())['data']
            log.debug(flowcell)

            if not flowcell:
                log.debug(">>> no flowcell")
                flowcell = self.minotourapi.create_flowcell(self.get_flowcell_id())

            is_barcoded = False  # TODO do we known this info at this moment? This can be determined from run info.

            has_fastq = True  # TODO do we known this info at this moment? This can be determined from run info
            log.debug(">>> before self.minotourapi.create_run")
            log.debug("self.sampleid.sample_id",self.sampleid.sample_id)
            createrun = self.minotourapi.create_run(self.sampleid.sample_id, runid, is_barcoded, has_fastq, flowcell, self.minion, self.runinformation.start_time.ToDatetime().strftime('%Y-%m-%d %H:%M:%S'))
            log.debug(">>> after self.minotourapi.create_run")

            # createrun = requests.post(self.args.full_host+'api/v1/runs/', headers=self.header, json={"run_name": self.status_summary['run_name'], "run_id": runid, "barcode": barcoded, "is_barcoded":is_barcoded, "minION":self.minion["url"]})

            if not createrun:
                log.error("Houston - we have a problem!")

            else:

                self.runidlink = createrun["url"]
                self.runid = createrun["id"]  # TODO is it id or runid?
                # self.runidlink = json.loads(createrun.text)["url"]
                # self.runid = json.loads(createrun.text)["id"]
                # self.create_flowcell(self.status_summary['flow_cell_id'])
                # self.create_flowcell_run()

        else:

            self.runidlink = run["url"]
            self.runid = run["id"]
        log.debug("***** self.runid", self.runid)

        try:
            ### I don't know what is happening here!
            #self.minotourapi.update_minion_run_stats()
            #self.minotourapi.update_minion_run_info()
            pass
        except Exception as err:
            log.debug("Problem minotourapi", err)
        log.debug("**** run stats updated")


    def run_live(self):
        """
        This function will update the run information to the server at a given rate during a live run.
        :return:
        """
        pass

    def run_stop(self):
        """
        This function will clean up when a run finishes.
        :return:
        """
        self.minotourapi.update_minion_event(self.minion, self.computer_name, "active")
        log.debug("run stop observed")


    def jobs_monitor(self):
        """
        This function will check the remote server for new jobs to be done.
        :return:
        """
        while True:
            log.debug("!!!!!!checking for jobs!!!!!!")
            jobs = self.minotourapi.get_minion_jobs(self.minion)
            log.debug(jobs)
            time.sleep(self.interval)
            for job in jobs:
                if job["job"] == "testmessage":
                    self.sendmessage(1,"minoTour is checking communication status with " + str(self.minion['name']) + ".")
                    self.minotourapi.complete_minion_job(self.minion,job)
                if job["job"] == "custommessage":
                    self.sendmessage(1,
                                     str(job["custom"]))
                    self.minotourapi.complete_minion_job(self.minion, job)
                if job["job"] == "stopminion":
                    if self.args.enable_remote:
                        self.rpc_connection.protocol.stop_protocol()
                        self.sendmessage(3,"MinoTour was used to remotely stop your run.")
                    self.minotourapi.complete_minion_job(self.minion, job)
                if job["job"] == "rename":
                    if self.args.enable_remote:
                        self.rpc_connection.protocol.set_sample_id(sample_id=job["custom"])
                        self.sendmessage(1,"MinoTour renamed your run to {}".format(job["custom"]))
                    self.minotourapi.complete_minion_job(self.minion, job)
                if job["job"] == "nameflowcell":
                    if self.args.enable_remote:
                        self.rpc_connection.device.set_user_specified_flow_cell_id(id=job["custom"])
                        self.sendmessage(1, "MinoTour renamed your flowcell to {}".format(job["custom"]))
                    self.minotourapi.complete_minion_job(self.minion, job)
                if job["job"] == "startminion":
                    if self.args.enable_remote:
                        print(job["custom"],"\n\n\n\n\n")
                        self.rpc_connection.protocol.start_protocol(identifier=job["custom"])
                        self.sendmessage(1,"MinoTour attempted to start a run on your device.")
                    self.minotourapi.complete_minion_job(self.minion, job)


    def minknow_command(self):
        """
        This function will recieve commands for a specific minION and handle the interaction.
        :return:
        """
        pass

    def get_flowcell_id(self):
        if len(self.flowcelldata['user_specified_flow_cell_id']) > 0:
            log.debug("We have a self named flowcell")
            return str(self.flowcelldata['user_specified_flow_cell_id'])
        else:
            log.debug("the flowcell id is fixed")
            return str(self.flowcelldata['flow_cell_id'])


    def flowcellmonitor(self):
        while True:
            flowcellinfo = self.rpc_connection.device.stream_flow_cell_info()
            for event in flowcellinfo:
                log.debug(event)
                self.flowcelldata = parsemessage(event)
                log.debug(self.get_flowcell_id())
                self.update_minion_status()


    def newhistogrammonitor(self):
        while True:
            histogram_stream = self.rpc_connection.statistics.stream_read_length_histogram(poll_time=60,wait_for_processing=True,read_length_type=0,bucket_value_type=1)

            try:
                for histogram_event in histogram_stream:
                    #print (parsemessage(histogram_event))
                    self.histogramdata = parsemessage(histogram_event)
                if not str(self.status).startswith("status: PROCESSING"):
                    break
            except:
                print ("Histogram Problem")
                pass
            time.sleep(self.interval)
            pass


    def newchannelstatemonitor(self):
        while True:
            channel_states = self.rpc_connection.data.get_channel_states(wait_for_processing=True, first_channel=1,
                                                                         last_channel=512)
            try:
                for state in channel_states:
                    for channel in state.channel_states:#print (state)
                        self.channelstates[int(channel.channel)]=channel.state_name
                if not str(self.status).startswith("status: PROCESSING"):
                    break
            except:
                pass
            time.sleep(self.interval)
            pass

    def dutytimemonitor(self):
        while True:
            log.debug("Duty Time Monitor Running", self.status)
            log.debug(str(self.status))
            while str(self.status).startswith("status: PROCESSING"):
                log.debug("fetching duty time")
                dutytime = self.rpc_connection.statistics.stream_duty_time(wait_for_processing=True,step=60)
                if self.args.verbose:
                    for duty in dutytime:
                        log.debug(duty)
            time.sleep(1)


    def runmonitor(self):
        while True:
            status_watcher = rpc.wrappers.StatusWatcher(self.rpc_connection)
            msgs = rpc.acquisition_service
            while True:
                for status in status_watcher.wait():
                    self.status = status
                    if str(self.status).startswith("status: STARTING"):
                        self.run_start()
                    if str(self.status).startswith("status: FINISHING"):
                        self.run_stop()
                    log.debug(status)

    def update_minion_status(self):
        #### This block of code will update live information about a minION
        ### We may not yet have a run to acquire - if so the acquisition_data will be empty.

        acquisition_data=dict()

        if len(self.acquisition_data)<1:
            acquisition_data['state']="No Run"
            currentscript = "Nothing Running"
        else:
            acquisition_data = self.acquisition_data
            currentscript = str(self.rpc_connection.protocol.get_run_info().protocol_id)

        if len(self.disk_space_info)<1:
            self.disk_space_info = json.loads(
                MessageToJson(self.rpc_connection.instance.get_disk_space_info(), preserving_proto_field_name=True,
                              including_default_value_fields=True))
            log.debug(self.disk_space_info)

        payload = {"minION": str(self.minion["url"]),
                   "minKNOW_status": acquisition_data['state'],
                   "minKNOW_current_script": currentscript,
                   #"minKNOW_sample_name": None,
                   "minKNOW_exp_script_purpose": str(self.rpc_connection.protocol.get_protocol_purpose()),
                   "minKNOW_flow_cell_id": self.get_flowcell_id(),
                   #"minKNOW_run_name": str(self.sampleid.sample_id),
                   #"minKNOW_hash_run_id": str(self.runinformation.run_id),
                   #"minKNOW_script_run_id": str(
                   #    self.rpc_connection.protocol.get_current_protocol_run().acquisition_run_ids[0]),
                   "minKNOW_real_sample_rate": int(
                       str(self.rpc_connection.device.get_sample_rate().sample_rate)),
                   "minKNOW_asic_id": self.flowcelldata['asic_id'],  # self.status_summary['asic_id'],
                   "minKNOW_total_drive_space": self.disk_space_info["filesystem_disk_space_info"][0]["bytes_capacity"],
                   "minKNOW_disk_space_till_shutdown": self.disk_space_info["filesystem_disk_space_info"][0]["bytes_when_alert_issued"],
                   "minKNOW_disk_available": self.disk_space_info["filesystem_disk_space_info"][0]["bytes_available"],
                   "minKNOW_warnings": self.disk_space_info["filesystem_disk_space_info"][0]["recommend_stop"],
                   # TODO work out what this should be! self.status_summary['recommend_alert'],
                   }
        try:
            payload["minKNOW_script_run_id"] = self.rpc_connection.protocol.get_current_protocol_run().acquisition_run_ids[0]
        except:
            pass
        if hasattr(self, 'sampleid'):
            payload["minKNOW_sample_name"]=str(self.sampleid.sample_id)
            payload["minKNOW_run_name"]=str(self.sampleid.sample_id)

        if hasattr(self, 'runinformation'):
            payload["minKNOW_hash_run_id"]=str(self.runinformation.run_id)

        if self.minIONstatus:  # i.e the minION status already exists

            self.minIONstatus = self.minotourapi.update_minion_status(payload, self.minion)

        else:

            self.minIONstatus = self.minotourapi.create_minion_status(payload, self.minion)

    def update_minion_stats (self):
        asictemp = self.temperaturedata.minion.asic_temperature.value
        heatsinktemp = self.temperaturedata.minion.heatsink_temperature.value
        biasvoltage = int(self.bias_voltage)
        voltage_val = int(self.bias_voltage)#todo this likely is wrong
        voltage_value = biasvoltage #todo check this = probably wrong
        yield_val = self.acquisition_data['yield_summary']['selected_events']
        read_count = self.acquisition_data['yield_summary']['read_count']
        channelpanda = pd.DataFrame.from_dict(self.channelstates, orient='index', dtype=None)
        channeldict=dict()
        channeldict["strand"]=0
        channeldict["adapter"]=0
        channeldict["good_single"]=0
        channeldict["pore"]=0
        try:
            channelpandastates = channelpanda.groupby([0,]).size()
            #print (channelpandastates)
            log.debug(channelpandastates)
            for state, value in channelpandastates.iteritems():
                log.debug(state, value)
            #    print (state,value)
                channeldict[state]=value
            #print ("\n\n\n\n\n\n")
            instrand = 0 #channeldict["strand"]+channeldict["adapter"]
            openpore = 0 #channeldict["good_single"]+channeldict["pore"]
            meanratio = 0 #todo work out if we can still do this
        except:
            meanratio=0
            instrand=0
            openpore=0
            pass

        # Capturing the histogram data from MinKNOW
        #print (self.runinformation)


        payload = {"minION": str(self.minion["url"]),
                   "run_id": self.runidlink,
                   "sample_time": str(datetime.datetime.now()),
                   "event_yield": yield_val,
                   "asic_temp": asictemp,
                   "heat_sink_temp": heatsinktemp,
                   "voltage_value": voltage_value,
                   "mean_ratio": meanratio,
                   "open_pore": openpore,
                   "in_strand": instrand,
                   "minKNOW_histogram_values": str(self.histogramdata["histogram_data"]["buckets"]),
                   "minKNOW_histogram_bin_width": self.histogramdata["histogram_data"]["width"],
                   "minKNOW_read_count": read_count
                   }
        for channel in channeldict:
            payload[str(channel)] = channeldict[channel]

        log.debug("This our new payload",payload)

        result = self.minotourapi.create_minion_statistic(payload,self.runid)

        log.debug("This is our result.", result)



    def runinfo(self):
        while True:
            log.debug("Checking run info")
            try:
                self.acquisition_data = parsemessage(self.rpc_connection.acquisition.get_acquisition_info())
            except:
                log.debug("No active run")
                self.acquisition_data ={}
            self.temperaturedata = self.rpc_connection.device.get_temperature()
            self.disk_space_info = json.loads(MessageToJson(self.rpc_connection.instance.get_disk_space_info(), preserving_proto_field_name=True, including_default_value_fields=True))
            if str(self.device_type).startswith("PROMETHION"):
                self.minion_settings = self.rpc_connection.promethion_device.get_device_settings()
            else:
                self.minion_settings = self.rpc_connection.minion_device.get_settings()
            self.bias_voltage = json.loads(MessageToJson(self.rpc_connection.device.get_bias_voltage(),preserving_proto_field_name=True,including_default_value_fields=True))['bias_voltage']

            try:
                self.runinfo_api = self.rpc_connection.protocol.get_run_info()
            except:
                log.debug("Run Info not yet known.")
            try:
                self.sampleid = self.rpc_connection.protocol.get_sample_id()
            except:
                log.debug("Sample ID not yet known.")
            log.debug("running update minion status")
            self.update_minion_status()
            if str(self.status).startswith("status: PROCESSING"):
                self.runinformation = self.rpc_connection.acquisition.get_current_acquisition_run()
                log.debug(self.runinformation)
                try:
                    log.debug("running update minion stats")
                    if hasattr(self, 'runid'):
                        self.update_minion_stats()
                except Exception as err:
                    log.error("Problem updating stats to device.", err)
                    pass
            try:
                log.debug(self.read_event_weighted_hist)
                log.debug(self.read_hist_bin_width)
            except:
                log.debug("Couldn't log histogram data.")
            time.sleep(self.interval)

    def sendmessage(self,severitylevel,message):
        self.rpc_connection.log.send_user_message(severity=severitylevel, user_message=message)

    def getmessages(self):
        while True:
            messages = self.rpc_connection.log.get_user_messages(include_old_messages=True)
            for message in messages:

                payload = {"minion": self.minion["url"],
                           "message": message.user_message,
                           "run": "",
                           "identifier": message.identifier,
                           "severity": message.severity,
                           "timestamp": message.time.ToDatetime().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3],
                           }

                if self.runidlink:
                    payload["run"] = self.runidlink

                messagein = self.minotourapi.create_message( payload, self.minion)


    def opened(self):
        log.debug("Connected to MinIONs")
        ##print "Trying \"engine_states\":\"1\",\"channel_states\":\"1\",\"multiplex_states\":\"1\""
        #self.send(json.dumps({'engine_states':'1','channel_states':'1','multiplex_states':'1','channel_info':'1'}))
        #self.send(json.dumps({'engine_states':'1','channel_states':'1','channel_info':'1'}))
        #self.send(json.dumps({'engine_states':'1','channel_states':'1'}))
        #self.send(json.dumps({'engine_states':'1'}))
        #self.send(json.dumps({'channel_info':'1','channel_states':'1'}))
        #self.send(transport.getvalue(), binary=True)

    def closed(self, code, reason="client disconnected for some reason"):
        log.info("socket",self.sock.fileno())
        log.info("Closed down", code, reason)

    def received_message(self, m):
        ## All we do here is grab the old read_hist_bin - we shouldn't need to do this anymore?
        ## This code is now deprecated
        if not m.is_binary:
            #print ("****************** Non binary message")
            json_object = json.loads(str(m))
            #print (json_object)
            try:
                for key in json_object:
                    #print (key)
                    #if str(key) == "engine_states":
                    #    print (json_object[key])
                    if str(key) == "statistics":
                        log.debug(json_object[key].keys())
                        if "read_event_count_weighted_hist_bin_width" in json_object[key]:
                            self.read_hist_bin_width = json_object[key]["read_event_count_weighted_hist_bin_width"]
                        else:
                            self.read_hist_bin_width = 0
                        if "read_event_count_weighted_hist" in json_object[key]:
                            self.read_event_weighted_hist = json_object[key]["read_event_count_weighted_hist"]
                        else:
                            self.read_event_weighted_hist = ""
                #print (json_object.keys())
            except:
                log.error("key error")

class MinknowConnectRPC():
    def __init__(self,minswip,args,header):
        self.args = args
        log.debug("initialising minknow connection via rpc")
        self.header = header
        self.minIONdict=dict() #A dictionary to store minION connection data.
        devicemonitorthread = threading.Thread(target=self.devicemonitor, args=())
        devicemonitorthread.daemon = True  # Daemonize thread
        devicemonitorthread.start()

    def devicemonitor(self):
        # -------------------------------------------------------------------------------
        # Connect to the running Manager instance:
        #
        # We can connect to minknow manager on port 9501.
        #
        channel = grpc.insecure_channel('localhost:9501')
        stub = manager_grpc.ManagerServiceStub(channel)
        while True:
            # Send the list request
            list_request = manager.ListDevicesRequest()
            response = stub.list_devices(list_request)
            for device in response.active:
                log.debug(device.name)
                deviceid = device.name
                if deviceid not in self.minIONdict:
                    self.minIONdict[deviceid] = dict()
                    self.minIONdict[deviceid]["state"] = "pending"
                if self.minIONdict[deviceid]["state"]!= "active":
                    self.minIONdict[deviceid]["grpc_port"] = device.ports.insecure_grpc
                    self.minIONdict[deviceid]["ws_longpoll_port"] = device.ports.json_websocket
                    log.debug(self.minIONdict[deviceid]["grpc_port"])
                    self.minIONdict[deviceid]["grpc_connection"] = rpc.Connection(
                        port=self.minIONdict[deviceid]["grpc_port"])
                    connectip = "ws://" + self.args.ip + ":" + str(self.minIONdict[deviceid]["ws_longpoll_port"]) + "/"

                    self.minIONdict[deviceid]["device_connection"] = DeviceConnect(connectip, self.args,
                                                                                   self.minIONdict[deviceid][
                                                                                       "grpc_connection"], self.header,
                                                                                   deviceid)
                    # self.minIONdict[deviceid]["legacydevicedata"] = DeviceConnectLegacy(connectip,self.args,self.minIONdict[deviceid]["grpc_connection"])
                    try:
                        self.minIONdict[deviceid]["device_connection"].connect()
                    except Exception as err:
                        log.error("Problem connecting to device.", err)
                    self.minIONdict[deviceid]["state"]="active"
            time.sleep(5)

    def minIONnumber(self):
        return len(self.minIONdict)

    def disconnect_nicely(self):
        for device in self.minIONdict:
            log.info("Disconnecting {} from the server.".format(device))
            self.minIONdict[device]["device_connection"].disconnect_nicely()
        log.info("Stopped successfully.")


def parsearguments():
    parser = \
        configargparse.ArgParser(description='interaction: A program to provide real time interaction for minION runs.')
    parser.add(
        '-ip',
        '--ip-address',
        type=str,
        dest='ip',
        required=True,
        default=None,
        help='The IP address of the minKNOW machine.',
    )
    parser.add(
        '-d',
        '--debug',
        action='store_true'
    )
    args = parser.parse_args()
    return args


def main():
    args = parsearguments()
    ##First thing to do is to try and connect to minKNOW to retrieve information on which minIONs are present.
    minwsip = "ws://" + args.ip + ":9500/"
    Minknow = MinknowConnect(minwsip, args)
    try:
        Minknow.connect()
    except Exception as err:
        log.error("Error", err)
        log.error(
            "We guess you have not got minKNOW running on your computer at the ip address specified. Please try again.")
        log.error("bye bye")
        sys.exit()

    while True:
        time.sleep(1)
        #print (Minknow.computer_name,Minknow.minknow_status,Minknow.minknow_version)
        #Minknow.reportinformation()

if __name__ == "__main__":
    main()



import gzip
import logging

from Bio import SeqIO

from minFQ.minotourapiclient import Runcollection


log = logging.getLogger(__name__)

def parsefastq(fastq, rundict, args, header):

    print('Inside parsefastq: {}, {}, {}'.format(fastq, rundict, args))
    counter = 0
    if fastq.endswith(".gz"):
        with gzip.open(fastq, "rt") as handle:
            for record in SeqIO.parse(handle, "fastq"):
                counter += 1
                args.fastqmessage = "processing read {}".format(counter)
                descriptiondict = parsedescription(record.description)
                if descriptiondict["runid"] not in rundict:
                    rundict[descriptiondict["runid"]] = Runcollection(args,header)
                    rundict[descriptiondict["runid"]].add_run(descriptiondict)
                rundict[descriptiondict["runid"]].add_read(record, descriptiondict, fastq)
    else:
        for record in SeqIO.parse(fastq, "fastq"):
            counter += 1
            args.fastqmessage = "processing read {}".format(counter)
            descriptiondict = parsedescription(record.description)
            if descriptiondict["runid"] not in rundict:
                rundict[descriptiondict["runid"]] = Runcollection(args, header)
                rundict[descriptiondict["runid"]].add_run(descriptiondict)
            rundict[descriptiondict["runid"]].add_read(record, descriptiondict,fastq)
    for runs in rundict:
        rundict[runs].commit_reads()


def parsedescription(description):
    descriptiondict = dict()
    descriptors = description.split(" ")
    del descriptors[0]
    for item in descriptors:
        bits = item.split("=")
        descriptiondict[bits[0]] = bits[1]
    return descriptiondict



### Functions for minknowconnection

def determinetype(minION):
    """
    :param minION:
    :return: devicetype,deviceid,portstring
    """
    #print (minION)
    devicetype = "unknown"
    deviceid = "unknown"
    if minION[1] == "M":
        devicetype = "MinION"
        deviceid=minION[1:8]
    elif minION[1] == "G":
        devicetype = "GridION"
        deviceid = minION[1:8]
    elif minION[1] =="1" or minION[1]=="2":
        devicetype = "PromethION"
        promvals = minION[1:9].split('-')
        if len(promvals[1])==2:
            deviceid = minION[1:8]
        else:
            deviceid = minION[1:10]
    return devicetype,deviceid  #,portstring

def parse_ports(ports,minIONname):
    """
    :param ports: string to parse representing the ports
    :return: list of ports
    """
    portset = list()
    rawports = list(map(lambda x: x, filter(lambda x: x, map(ord, ports))))
    chomplen = len(minIONname)+4
    rawports = rawports[chomplen:-2]
    rawvalues = list()
    offset = list()
    for index,port in enumerate(rawports):
        if index%3==0:
            try:
                rawvalues.append(port)
                offset.append(rawports[(index+1)])
            except:
                pass
    for index,rawvalue in enumerate(rawvalues):
        try:
            off = offset[index]
            if off == 62:
                correction = 192
            elif off == 63:
                correction = 64
            else:
                correction = 0
            port = rawvalue - correction + 8000
            portset.append(port)
        except:
            pass
    return portset



import os
import sys
import datetime
import configargparse
import validators

from ont_fast5_api import fast5_file, multi_fast5

from minFQ.minotourapi import MinotourAPI

from tqdm import tqdm


def main():
    parser = \
        configargparse.ArgParser(
            description='minFQ: A program to analyse minION fastq files in real-time or post-run and monitor the activity of MinKNOW.'
        )

    parser.add(
        '-w',
        '--watch-dir',
        type=str,
        required=True,
        default=None,
        help='The path to the folders containing fast5 reads to analyse - e.g. C:\\data\\minion\\downloads (for windows).',
        dest='watchdir'
    )

    parser.add(
        '-k',
        '--key',
        type=str,
        required=True,
        default=None,
        help='The api key for uploading data.',
        dest='api_key',
    )

    parser.add(
        '-p',
        '--port',
        type=int,
        # required=True,
        default=80,
        help='The port number for the local server.',
        dest='port_number',
    )

    parser.add(
        '-hn',
        '--hostname',
        type=str,
        required=True,
        default='127.0.0.1',
        help='The host name you are loading data too.',
        dest='host_name',
    )

    args = parser.parse_args()




    minION = set()
    flowcelldict = dict()
    rundict = dict()



    print('Searching for files in {}'.format(os.path.abspath(args.watchdir)))
    last_seen=""
    for root, subdirs, files in tqdm(os.walk(args.watchdir)):
        files.sort()
        for file in tqdm(files):
            if file.endswith("fast5"):
                if file.split('_read_')[0] != last_seen:
                    last_seen = file.split('_read_')[0]
                    try:
                        try:
                            myfile  = fast5_file.Fast5File(os.path.join(root,file))

                        except:
                            mymultifile = multi_fast5.MultiFast5File(os.path.join(root,file))
                            myid = mymultifile.get_read_ids()[0]
                            myfile = mymultifile.get_read(myid)
                        minIONid = myfile.get_tracking_id()['device_id']
                        minION.add(minIONid)
                        flowcell = myfile.get_tracking_id()['flow_cell_id']
                        if len(flowcell) < 1:
                            flowcell = input('Enter a flowcell name: ')
                        sample_id = myfile.get_tracking_id()['sample_id']
                        run_id = myfile.get_tracking_id()['run_id']
                        if sample_id not in flowcelldict.keys():
                            flowcelldict[sample_id]=dict()
                        if flowcell not in flowcelldict[sample_id].keys():
                            flowcelldict[sample_id][flowcell] = dict()
                        if run_id not in flowcelldict[sample_id][flowcell].keys():
                            flowcelldict[sample_id][flowcell][run_id]=dict()
                            flowcelldict[sample_id][flowcell][run_id]["tracking_id"]=myfile.get_tracking_id()
                            flowcelldict[sample_id][flowcell][run_id]["context_tags"]=myfile.get_context_tags()
                    except:
                        print ("Non fast5file seen.")
    print ("Adding new runs.")

    header = {
        'Authorization': 'Token {}'.format(args.api_key),
        'Content-Type': 'application/json'
    }
    minotourapi = MinotourAPI(args.host_name,args.port_number, header)
    for minIONs in minION:
        minion = minotourapi.get_minion_by_name(minIONs)
        if not minion:
            minion = minotourapi.create_minion(minIONs)
        minion = minion
    for sample_id in tqdm(flowcelldict):
        for flowcellid in flowcelldict[sample_id].keys():
            flowcell = minotourapi.get_flowcell_by_name(flowcellid)
            if len(flowcell["data"])==0:
                create_flowcell = minotourapi.create_flowcell(flowcellid)
                flowcell = minotourapi.get_flowcell_by_name(flowcellid)
            for run_id in flowcelldict[sample_id][flowcellid]:
                run = minotourapi.get_run_by_runid(run_id)
                if not run:
                    '''print (flowcell)'''
                    is_barcoded = False  # TODO do we known this info at this moment? This can be determined from run info.
                    has_fastq = True  # TODO do we known this info at this moment? This can be determined from run info
                    run = minotourapi.create_run(sample_id, run_id, is_barcoded, has_fastq,
                                                            flowcell['data'], minion,
                                                            flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['exp_start_time'])

                minion = minotourapi.get_minion_by_name(flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['device_id'])

                try: 
                    sample_rate =int(str(flowcelldict[sample_id][flowcellid][run_id]['context_tags']['sample_frequency']))
                except:
                    sample_rate = 0

                payload = {
                    "minION": str(minion["url"]),
                    "minKNOW_current_script": str(""),
                    "minKNOW_sample_name": str(sample_id),
                    "minKNOW_exp_script_purpose": str(flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['exp_script_purpose']),
                    "minKNOW_flow_cell_id": flowcellid,
                    "minKNOW_run_name": str(""),
                    "run_id": str(run['url']),
                    "minKNOW_version": str(flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['version']),
                    "minKNOW_hash_run_id": str(run_id),
                    "minKNOW_script_run_id": str(""),
                    "minKNOW_real_sample_rate": sample_rate,
                    "minKNOW_asic_id": flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['asic_id'],
                    "minKNOW_start_time": flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['exp_start_time'],
                    "minKNOW_computer": str(flowcelldict[sample_id][flowcellid][run_id]['tracking_id']['hostname']),
                }
                try:
                    payload['flowcell_type']=str(flowcelldict[sample_id][flowcellid][run_id]['context_tags']['flowcell_type'])
                except:
                    payload['flowcell_type']="unknown"
                try:
                    payload['sequencing_kit']=str(flowcelldict[sample_id][flowcellid][run_id]['context_tags']['sequencing_kit'])
                except:
                    payload['sequencing_kit']="unknown"

                updateruninfo = minotourapi.update_minion_run_info(payload, run['id'])

    print ("Metadata Uploaded to MinoTour.")
